/**
 * PETALS: PETALS Services Platform
 * Copyright (C) 2005 EBM WebSourcing
 *
 * This library is free software; you can redistribute it and/or
 * modify it under the terms of the GNU Lesser General Public
 * License as published by the Free Software Foundation; either
 * version 2.1 of the License, or any later version.
 *
 * This library is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 * Lesser General Public License for more details.
 *
 * You should have received a copy of the GNU Lesser General Public
 * License along with this library; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307
 * USA.
 *
 * Initial developer(s): EBM WebSourcing
 * --------------------------------------------------------------------------
 * $Id: AutoLoaderServiceImpl.java,v 1.2 2006/04/07 10:24:27 alouis Exp $
 * --------------------------------------------------------------------------
 */

package org.ow2.petals.jbi.management.autoload;

import java.io.File;
import java.io.FileNotFoundException;
import java.io.IOException;
import java.io.PrintWriter;
import java.net.URI;
import java.util.ArrayList;
import java.util.List;
import java.util.Timer;

import javax.management.MBeanServer;
import javax.management.ObjectName;

import org.apache.commons.io.FileUtils;
import org.objectweb.fractal.fraclet.annotation.annotations.FractalComponent;
import org.objectweb.fractal.fraclet.annotation.annotations.Interface;
import org.objectweb.fractal.fraclet.annotation.annotations.LifeCycle;
import org.objectweb.fractal.fraclet.annotation.annotations.Monolog;
import org.objectweb.fractal.fraclet.annotation.annotations.Provides;
import org.objectweb.fractal.fraclet.annotation.annotations.Requires;
import org.objectweb.fractal.fraclet.annotation.annotations.type.LifeCycleType;
import org.objectweb.util.monolog.api.Logger;
import org.ow2.petals.jbi.descriptor.JBIDescriptor;
import org.ow2.petals.jbi.management.admin.AdminServiceMBean;
import org.ow2.petals.jbi.management.deployment.DeploymentServiceMBean;
import org.ow2.petals.jbi.management.installation.InstallationServiceMBean;
import org.ow2.petals.jbi.management.util.PackageHandler;
import org.ow2.petals.jbi.management.util.XMLResult;
import org.ow2.petals.platform.repository.RepositoryService;
import org.ow2.petals.platform.systemstate.ComponentState;
import org.ow2.petals.platform.systemstate.ServiceAssemblyState;
import org.ow2.petals.platform.systemstate.SharedLibraryState;
import org.ow2.petals.platform.systemstate.SystemStateService;
import org.ow2.petals.util.LoggingUtil;
import org.ow2.petals.util.SystemUtil;

/**
 * This class is used to load automatically components which are placed in the
 * "autoLoad" directory. It scans the directory periodically and check if new
 * components are presents. The {@ link InstallationServiceMBean} is called for each of
 * those components. The whole process of installation is done (install + start)
 * After being threat, the archives of the components are removed. A file
 * "fileName.success" will be created if success A file "fileName.failed" will
 * be created otherwise, containing the related exception
 *
 * @author alouis
 *
 */
@FractalComponent
@Provides(interfaces = @Interface(name = "service", signature = org.ow2.petals.jbi.management.autoload.AutoLoaderService.class))
public class AutoLoaderServiceImpl implements AutoLoaderService {

    private static final String install = "install";

    private static final String installed = "installed";

    private static final String uninstalled = "uninstalled";

    private static final String work = "work";

    private static final long scanPeriod = 4000L;

    @Monolog(name = "logger")
    protected Logger logger;

    protected LoggingUtil log;

    protected static File installDirectory;

    protected static File installedDirectory;

    protected static File uninstalledDirectory;

    protected static File workDirectory;

    protected Timer installDirectoryScanTimer;

    protected Timer installedDirectoryScanTimer;

    /**
     * Package Handler
     */
    protected PackageHandler packageHandler;


    @Requires(name = "deployment", signature = org.ow2.petals.jbi.management.deployment.DeploymentServiceMBean.class)
    protected DeploymentServiceMBean deploymentService;

    @Requires(name = "installation", signature = org.ow2.petals.jbi.management.installation.InstallationServiceMBean.class)
    protected InstallationServiceMBean installationService;

    @Requires(name = "admin", signature = org.ow2.petals.jbi.management.admin.AdminServiceMBean.class)
    protected AdminServiceMBean adminService;

    /**
     * Container :: System state service
     */
    @Requires(name = "systemstate", signature = org.ow2.petals.platform.systemstate.SystemStateService.class)
    protected SystemStateService systemStateService;

    /**
     * Platform Component :: Repository Service
     */
    @Requires(name = "repository", signature = org.ow2.petals.platform.repository.RepositoryService.class)
    protected RepositoryService repositorySrv;


    public File getInstalledDirectory() {
        return installedDirectory;
    }

    public File getUninstalledDirectory() {
        return uninstalledDirectory;
    }

    public File getWorkDirectory() {
        return workDirectory;
    }

    /**
     * init the autoloader : init the directories
     */
    @LifeCycle(on = LifeCycleType.START)
    public void start() throws IOException {
        log = new LoggingUtil(logger);
        log.call();
        installDirectory = new File(SystemUtil.getPetalsInstallDirectory()
            .getAbsolutePath()
            + File.separator + install);
        installedDirectory = new File(SystemUtil.getPetalsInstallDirectory()
            .getAbsolutePath()
            + File.separator + installed);
        uninstalledDirectory = new File(SystemUtil.getPetalsInstallDirectory()
            .getAbsolutePath()
            + File.separator + uninstalled);
        workDirectory = new File(SystemUtil.getPetalsInstallDirectory()
            .getAbsolutePath()
            + File.separator + work);

        checkValidityOfDirectory(installDirectory, false);
        checkValidityOfDirectory(installedDirectory, false);
        checkValidityOfDirectory(uninstalledDirectory, false);
        checkValidityOfDirectory(workDirectory, true);

        // old start method
        installDirectoryScanTimer = new Timer();
        installedDirectoryScanTimer = new Timer();
        InstallDirectoryScanner installScan = new InstallDirectoryScanner(this,
            installDirectory);
        InstalledDirectoryScanner installedScan = new InstalledDirectoryScanner(
            this, installedDirectory);
        /*
         * Initialize package handler
         */
        packageHandler = new PackageHandler(logger, repositorySrv);

        installDirectoryScanTimer.schedule(installScan, scanPeriod, scanPeriod);
        installedDirectoryScanTimer.schedule(installedScan, scanPeriod,
            scanPeriod);
    }

    /**
     * Stop the autoload of the components. TODO : synchronization, does the
     * autoload must wait before the end of an installation before closing?
     */
    @LifeCycle(on = LifeCycleType.STOP)
    public void stop() {
        log.call();

        if (installDirectoryScanTimer != null) {
            installDirectoryScanTimer.cancel();
        }
        if (installedDirectoryScanTimer != null) {
            installedDirectoryScanTimer.cancel();
        }
    }

    /**
     * friendly method for the directoryscanner
     *
     * @param newInstallers
     */
    public void install(File[] newInstallers) {
        log.call();

        if (newInstallers != null) {
            for (int i = 0; i < newInstallers.length; i++) {
                File file = newInstallers[i];
                try {
                    /*
                     * Copy archive file to work dir
                     */
                    File workArchive = new File(getWorkDirectory(), file
                        .getName());
                    FileUtils.copyFile(file, workArchive);
                    FileUtils.forceDelete(file);

                    /*
                     * Load JBI descriptor here
                     */
                    JBIDescriptor descriptor = packageHandler
                        .loadDescriptor(workArchive.toURI());
                    /*
                     * Perform the right task : install, share or deploy
                     */
                    if (descriptor.getComponent() != null) {
                        performInstall(workArchive);
                    } else if (descriptor.getSharedLibrary() != null) {
                        performShare(workArchive);
                    } else if (descriptor.getServiceAssembly() != null) {
                        performDeploy(workArchive);
                    } else {
                        throw new Exception("Installation of entity failed : "
                            + workArchive.getCanonicalPath());
                    }
                    FileUtils.forceDelete(workArchive);

                } catch (Throwable e) {
                    log.error("Error during install", e);
                }
            }
        }
    }

    /**
     * friendly method for the directoryscanner
     *
     * @param newInstallers
     */
    public void uninstall(List<File> filesToUninstall) {
        log.call();
        /*
         * Retrieve all installed entities
         */
        List<ComponentState> compStates = systemStateService
            .retrieveAllComponentStates();
        List<SharedLibraryState> slStates = systemStateService
            .retrieveAllSharedLibraryStates();
        List<ServiceAssemblyState> saStates = systemStateService
            .retrieveAllServiceAssemblyStates();

        /*
         * Check if one of the files to uninstall match one of the archiveUrl of
         * the installed entities. If yes, this entity must be uninstalled.
         */
        List<String> compToUninstall = new ArrayList<String>();
        List<String> slToUninstall = new ArrayList<String>();
        List<String> saToUninstall = new ArrayList<String>();
        for (ComponentState state : compStates) {
            URI archiveURI = packageHandler.processAndGetPackageURI(state
                .getArchiveURL(), false);
            File archiveFile = new File(archiveURI);
            for (File file : filesToUninstall) {
                if (file.equals(archiveFile)) {
                    compToUninstall.add(state.getName());
                }
            }
        }
        for (SharedLibraryState state : slStates) {
            URI archiveURI = packageHandler.processAndGetPackageURI(state
                .getArchiveURL(), false);
            File archiveFile = new File(archiveURI);
            for (File file : filesToUninstall) {
                if (file.equals(archiveFile)) {
                    slToUninstall.add(state.getName());
                }
            }
        }
        for (ServiceAssemblyState state : saStates) {
            URI archiveURI = packageHandler.processAndGetPackageURI(state
                .getArchiveURL(), false);
            File archiveFile = new File(archiveURI);
            for (File file : filesToUninstall) {
                if (file.equals(archiveFile)) {
                    saToUninstall.add(state.getName());
                }
            }
        }

        /*
         * Perform entity uninstallation
         */
        for (String name : saToUninstall) {
            performSAUninstall(name);
        }
        for (String name : compToUninstall) {
            performCompUninstall(name);
        }
        for (String name : slToUninstall) {
            performSLUninstall(name);
        }

    }

    /**
     * Check if the specified File is a directory and is created. If it does not
     * exist, create it.
     *
     * @param dir
     * @param clean
     *            delete directory content or not
     */
    private void checkValidityOfDirectory(File dir, boolean clean)
        throws IOException {
        if (dir != null) {
            // parameter is not null
            if (!dir.exists()) {
                // directory does not exist, create it
                if (!dir.mkdirs()) {
                    // creation failed
                    throw new IOException(
                        "can not create the following directory : "
                            + dir.getCanonicalPath());
                }
            } else {
                // file already exist, check it is a directory
                if (!dir.isDirectory()) {
                    // file is not a directory
                    throw new IOException(dir.getCanonicalPath()
                        + " already exists and is not a directory.");
                } else {
                    // clean the directory
                    if (clean) {
                    	for (File file : dir.listFiles()) {
							file.delete();
						}
                    }
                }
            }
        } else {
            throw new FileNotFoundException("the specified directory is null");
        }
    }

    /**
     * Uninstall the specified component.
     *
     * @param name
     */
    protected void performCompUninstall(String name) {
        installationService.forceUnloadComponentInstaller(name);
    }

    /**
     * Deploy the specified Archive in the JBI container and start it . Any
     * exception will cause the abortion of the process. A file
     * "fileName.success" will be created if success A file "fileName.failed"
     * will be created otherwise, containing the related exception
     *
     * @param f
     */
    protected void performDeploy(File f) {
        log.call();

        if (f != null && f.exists()) {

            try {

                // TODO Find a better way of managing URI
                // String absPath = f.getAbsolutePath().replaceAll(" ", "%20");
                // String archiveURI;
                // if (absPath.startsWith("/")) {
                // archiveURI = "file://" + absPath;
                // } else {
                // archiveURI = "file:///" + absPath;
                // }

                String xml = deploymentService.deploy(f.toURI().toString());

                if (xml.indexOf(XMLResult.TaskResult.FAILED.toString()) > -1) {
                    log.warning("Problem while deploying a service assembly\n"
                        + xml);
                } else {
                    String saName = "";
                    saName = xml.substring(xml.indexOf("<loc-param>") + 11, xml
                        .indexOf("</loc-param>"));
                    deploymentService.start(saName);

                }
            } catch (Throwable e) {
                try {
                    createFailureFile(f, e);
                    log.error("Error occured during auto deployment", e);
                } catch (IOException e1) {
                    log.error(
                        "Error occured during failure report file creation.",
                        e1);
                }
            }
        }
    }

    /**
     * Install the specified Archive in the JBI container and start it . Any
     * exception will cause the abortion of the process. A file
     * "fileName.success" will be created if success A file "fileName.failed"
     * will be created otherwise, containing the related exception
     *
     * @param f
     */
    protected void performInstall(File f) {
        log.call();

        if (f != null && f.exists()) {

            try {

                MBeanServer jmxServer = adminService.getMBeanServer();

                // TODO Find a better way of managing URI
                // String absPath = f.getAbsolutePath().replaceAll(" ", "%20");
                // String archiveURI;
                // if (absPath.startsWith("/")) {
                // archiveURI = "file://" + absPath;
                // } else {
                // archiveURI = "file:///" + absPath;
                // }

                ObjectName newInstallerName = installationService
                    .loadNewInstaller(f.toURI().toString());
                if (!newInstallerName.getDomain().equals("error")) {
                    Object[] nullObjects = new Object[0];
                    String[] nullStrings = new String[0];

                    Object result = jmxServer.invoke(newInstallerName,
                        "install", nullObjects, nullStrings);

                    if (result instanceof ObjectName) {
                        ObjectName componentName = (ObjectName) result;

                        jmxServer.invoke(componentName, "start", nullObjects,
                            nullStrings);

                    } else {
                        throw new Exception("error during installation");
                    }
                } else {
                    throw new Exception("Component can't be autoinstalled");
                }
            } catch (Throwable e) {
                try {

                    createFailureFile(f, e);
                    log.error("Error occured during auto installation.", e);
                } catch (IOException e1) {
                    log.error(
                        "Error occured during failure report file creation.",
                        e1);
                }
            }
        }
    }

    private void createFailureFile(File f, Throwable e) throws IOException,
        FileNotFoundException {
        File failed = new File(workDirectory, f.getName() + ".failed");

        failed.createNewFile();
        PrintWriter w = null;
        w = new PrintWriter(failed);
        e.printStackTrace(w);
        w.flush();
        w.close();
    }

    /**
     * Undeploy the specified service assembly. isToBeDeleted is setted to true !
     *
     * @param name
     */
    protected void performSAUninstall(String name) {
        deploymentService.forceUndeployServiceAssembly(name);
    }

    /**
     * Install the specified shared library in the JBI container Any exception
     * will cause the abortion of the process. A file "fileName.success" will be
     * created if success A file "fileName.failed" will be created otherwise,
     * containing the related exception
     *
     * @param f
     */
    protected void performShare(File f) {
        log.call();

        if (f != null && f.exists()) {

            try {

                // TODO Find a better way of managing URI
                // String absPath = f.getAbsolutePath().replaceAll(" ", "%20");
                // String archiveURI;
                // if (absPath.startsWith("/")) {
                // archiveURI = "file://" + absPath;
                // } else {
                // archiveURI = "file:///" + absPath;
                // }

                installationService.installSharedLibrary(f.toURI().toString());

            } catch (Throwable e) {
                try {

                    createFailureFile(f, e);
                    log.error(
                        "Error occured during auto share lib installation.", e);
                } catch (IOException e1) {
                    log.error(
                        "Error occured during failure report file creation.",
                        e1);
                }
            }
        }
    }

    /**
     * Uninstall the specified shared lib.
     *
     * @param name
     */
    protected void performSLUninstall(String name) {
        installationService.uninstallSharedLibrary(name);
    }

}
