/**
 * PETALS - PETALS Services Platform.
 * Copyright (c) 2005 Ebm Websourcing
 *
 * This library is free software; you can redistribute it and/or
 * modify it under the terms of the GNU Lesser General Public
 * License as published by the Free Software Foundation; either
 * version 2.1 of the License, or (at your option) any later version.
 * This library is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 * Lesser General Public License for more details.
 *
 * You should have received a copy of the GNU Lesser General Public
 * License along with this library; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
 *
 * -------------------------------------------------------------------------
 * $Id$
 * -------------------------------------------------------------------------
 */
package org.ow2.petals.jbi.management.installation;

import java.net.MalformedURLException;
import java.net.URI;
import java.net.URL;
import java.util.ArrayList;
import java.util.HashMap;
import java.util.List;

import javax.jbi.JBIException;
import javax.management.ObjectName;
import javax.naming.InitialContext;

import org.objectweb.fractal.fraclet.annotation.annotations.FractalComponent;
import org.objectweb.fractal.fraclet.annotation.annotations.Interface;
import org.objectweb.fractal.fraclet.annotation.annotations.LifeCycle;
import org.objectweb.fractal.fraclet.annotation.annotations.Monolog;
import org.objectweb.fractal.fraclet.annotation.annotations.Provides;
import org.objectweb.fractal.fraclet.annotation.annotations.Requires;
import org.objectweb.fractal.fraclet.annotation.annotations.type.LifeCycleType;
import org.objectweb.util.monolog.api.Logger;
import org.ow2.petals.PetalsException;
import org.ow2.petals.container.ContainerService;
import org.ow2.petals.container.lifecycle.ComponentLifeCycle;
import org.ow2.petals.container.lifecycle.Installer;
import org.ow2.petals.jbi.management.admin.AdminServiceMBean;
import org.ow2.petals.jbi.management.autoload.AutoLoaderService;
import org.ow2.petals.jbi.management.task.TaskContextConstants;
import org.ow2.petals.jbi.management.task.installation.InstallationContextConstants;
import org.ow2.petals.jbi.management.task.installation.InstallationTaskFactory;
import org.ow2.petals.jbi.management.util.PackageHandler;
import org.ow2.petals.jbi.messaging.routing.Router;
import org.ow2.petals.platform.classloader.ClassLoaderService;
import org.ow2.petals.platform.repository.RepositoryService;
import org.ow2.petals.platform.systemstate.SystemStateService;
import org.ow2.petals.processor.TaskProcessor;
import org.ow2.petals.util.LoggingUtil;
import org.ow2.petals.util.ParameterCheckHelper;

/**
 * JBI installation service implementation.
 *
 * @version $Rev: 154 $ $Date: 2006-03-27 17:30:10 +0200 (lun., 27 mars 2006) $
 * @since Petals 1.0
 * @author <a href="mailto:rmarins@fossilec.com">Rafael Marins</a>
 * @author ofabre - EBM Websourcing
 */
@FractalComponent
@Provides(interfaces = @Interface(name = "service", signature = org.ow2.petals.jbi.management.installation.InstallationServiceMBean.class))
public class InstallationServiceImpl implements InstallationServiceMBean {

	/**
	 * logger wrapper
	 */
	protected LoggingUtil log;

	/**
	 * Monolog Logger instance.
	 */
	@Monolog(name = "logger")
	protected Logger logger;

	/**
	 * Package Handler.
	 */
	protected PackageHandler packageHandler;

    /**
     * Communication :: JMX Agent Admin.
     */
    @Requires(name = "admin", signature = org.ow2.petals.jbi.management.admin.AdminServiceMBean.class)
    protected AdminServiceMBean adminService;

    /**
     * JBI management :: Autoloader Service.
     */
    @Requires(name = "autoloader", signature = org.ow2.petals.jbi.management.autoload.AutoLoaderService.class)
    protected AutoLoaderService autoLoaderService;

	/**
	 * JBI messaging :: Router Service.
	 */
	@Requires(name = "router", signature = org.ow2.petals.jbi.messaging.routing.Router.class)
	protected Router router;

    /**
     * Platform :: Loader Service.
     */
    @Requires(name = "classloader", signature = org.ow2.petals.platform.classloader.ClassLoaderService.class)
    protected ClassLoaderService loaderService;

    /**
     * Platform :: Repository Service.
     */
    @Requires(name = "repository", signature = org.ow2.petals.platform.repository.RepositoryService.class)
    protected RepositoryService repositoryService;

    /**
     * Platform :: SystemState Service.
     */
    @Requires(name = "systemstate", signature = org.ow2.petals.platform.systemstate.SystemStateService.class)
    protected SystemStateService systemStateService;

	/**
	 * Container :: Container Service.
	 */
	@Requires(name = "container", signature = org.ow2.petals.container.ContainerService.class)
	protected ContainerService container;

	protected InitialContext componentInitialContext;

	protected InstallationTaskFactory installationTaskFactory;

	/**
	 * Default constructor.
	 */
	public InstallationServiceImpl() {

	}

	/**
	 * @see javax.jbi.management.InstallationServiceMBean#installSharedLibrary(java.lang.String)
	 *      Return ERROR_INSTALL_SHARED_LIBRARY if shared library can't be
	 *      loaded correctly. It's an implementation choice !
	 */
	public synchronized String installSharedLibrary(String slZipURL) {
		log.start();
		ParameterCheckHelper.isNullParameterWithLog(slZipURL, "slZipURL", log);
		ParameterCheckHelper.isEmptyParameterWithLog(slZipURL, "slZipURL", log);
		try {
			new URL(slZipURL);
		} catch (MalformedURLException e) {
			throw new IllegalArgumentException("'" + slZipURL
					+ "' is not a valid URL");
		}

		/*
		 * Transform URL into URI
		 */
		URI archiveURI = packageHandler
				.processAndGetPackageURI(slZipURL, false);

		String msg;
		HashMap<String, Object> installationContext = null;

		try {

			/*
			 * Create installation execution context
			 */
			installationContext = createInstallationContext(archiveURI);

			/*
			 * Create installation processor and fill with tasks
			 */
			TaskProcessor processor = createSLInstallationProcessor(
					installationContext, archiveURI);

			/*
			 * Launch the process
			 */
			if (!processor.process()) {
				msg = "An error occured during the installation "
						+ "of the following shared lib: " + slZipURL;
				throw new Exception(msg);
			}
			log.info("Shared library succesfully installed: "
					+ (String) installationContext
							.get(TaskContextConstants.ENTITY_NAME));

		} catch (Throwable e) {
			msg = "Could not install the shared library: " + slZipURL;
			log.error(msg, e);

			throw new RuntimeException(msg);
		}

		log.end();

		return (String) installationContext
				.get(TaskContextConstants.ENTITY_NAME);
	}

	/**
	 * @see javax.jbi.management.InstallationServiceMBean#loadInstaller(java.lang.String)
	 */
	public ObjectName loadInstaller(String componentName) {
		ParameterCheckHelper.isNullParameterWithLog(componentName,
				"componentName", log);
		ParameterCheckHelper.isEmptyParameterWithLog(componentName,
				"componentName", log);
		ObjectName result = adminService.getInstallerByName(componentName);

		// Call init on bootstrap
		Installer installer = adminService.getInstallerMBeanByName(result);
		if (installer != null) {
			try {
				installer.initBootstrap(false);
			} catch (JBIException e) {
				String msg = "Can't initialize Bootstrap !";
				log.error(msg, e);
				throw new RuntimeException(msg);
			}
		}

		return result;
	}

	/**
	 * @see javax.jbi.management.InstallationServiceMBean#loadNewInstaller(java.lang.String)
	 *      Return a specific ObjectName if component can't be installed
	 *      correctly. It's an implementation choice !
	 * @throws RuntimeException
	 *             if an error occurs during installation process. We have
	 *             chosen this type of exception because, in the specification,
	 *             this method doesn't return an exception
	 */
	public ObjectName loadNewInstaller(String installZipURL) {
		log.start();
		ParameterCheckHelper.isNullParameterWithLog(installZipURL,
				"installZipURL", log);
		ParameterCheckHelper.isEmptyParameterWithLog(installZipURL,
				"installZipURL", log);
		try {
			new URL(installZipURL);
		} catch (Throwable e) {
			throw new IllegalArgumentException(
					"installZipUrl must be a legal url", e);
		}
		String msg = null;

		/*
		 * Transform URL into URI
		 */
		URI archiveURI = packageHandler.processAndGetPackageURI(installZipURL,
				false);

		HashMap<String, Object> installationContext = null;

		try {

			/*
			 * Create installation execution context
			 */
			installationContext = createInstallationContext(archiveURI);

			/*
			 * Create installation processor and fill with tasks
			 */
			TaskProcessor processor = createComponentInstallationProcessor(
					installationContext, archiveURI);

			/*
			 * Launch the process
			 */
			if (!processor.process()) {
				msg = "An error occured during the installation "
						+ "of the following component: " + installZipURL;
				throw new Exception(msg);
			}
			log.info("Component succesfully installed: "
					+ (String) installationContext
							.get(TaskContextConstants.ENTITY_NAME));

		} catch (Throwable e) {
			msg = "Component can't be installed: " + installZipURL;
			log.error(msg, e);

			throw new RuntimeException(msg);
		}
		log.end();

		return (ObjectName) installationContext
				.get(InstallationContextConstants.INSTALLER_OBJECT_NAME);
	}

	/**
	 * Create the installation context
	 *
	 * @param zipUrl
	 *            String containing the location URL of the Component or Shared
	 *            lib ZIP file
	 * @return the Installation Context HashMap
	 */
	private HashMap<String, Object> createInstallationContext(URI archiveURI) {
		HashMap<String, Object> installationContext = new HashMap<String, Object>();
		installationContext.put(TaskContextConstants.ARCHIVE_URI, archiveURI);
		return installationContext;
	}

	/**
	 * Create the installation processor and add all required task for the
	 * complete installation of a component
	 *
	 * @param installationContext
	 *            the installation context
	 * @param archiveURI
	 *            the component archive URI
	 * @return the filled installation processor
	 */
	private TaskProcessor createComponentInstallationProcessor(
			HashMap<String, Object> installationContext, URI archiveURI) {
		TaskProcessor processor = new TaskProcessor(installationContext, log);

		String scheme = archiveURI.getScheme();
		if (scheme.equalsIgnoreCase("http") || scheme.equalsIgnoreCase("ftp")) {
			processor.addTask(installationTaskFactory.getDownloadTask());
			processor.addTask(installationTaskFactory.getExtractTask());
		} else if (scheme.equalsIgnoreCase("file")) {
			if (!packageHandler.isDirectory(archiveURI)) {
				processor.addTask(installationTaskFactory.getExtractTask());
			} else {
				processor.addTask(installationTaskFactory
						.getCopyArchiveToWorkDirTask());
			}
		} else {
			throw new IllegalArgumentException("URL protocol not supported:"
					+ scheme);
		}

		processor.addTask(installationTaskFactory
				.getComponentCheckPackageTask());

		processor.addTask(installationTaskFactory
				.getCopyComponentPackageToRepositoryTask());
		processor.addTask(installationTaskFactory
				.getCheckInstalledSLForCompTask());
		processor.addTask(installationTaskFactory
				.getCreateAndRegisterInstallerMBeanTask());

		processor.addTask(installationTaskFactory
				.getCreateComponentStateHolderTask());
		if (scheme.equalsIgnoreCase("file")
				&& !packageHandler.isDirectory(archiveURI)) {
			processor.addTask(installationTaskFactory
					.getCopyArchiveToInstalledDirTask());
		}

		processor.addTask(installationTaskFactory.getSuccessTask());

		return processor;
	}

	/**
	 * Create the installation processor and add all required task for the
	 * complete installation of a shared lib
	 *
	 * @param installationContext
	 *            the installation context
	 * @param archiveURI
	 *            the sl archive URI
	 * @return the filled installation processor
	 */
	private TaskProcessor createSLInstallationProcessor(
			HashMap<String, Object> installationContext, URI archiveURI) {
		TaskProcessor processor = new TaskProcessor(installationContext, log);

		String scheme = archiveURI.getScheme();
		if (scheme.equalsIgnoreCase("http") || scheme.equalsIgnoreCase("ftp")) {
			processor.addTask(installationTaskFactory.getDownloadTask());
			processor.addTask(installationTaskFactory.getExtractTask());
		} else if (scheme.equalsIgnoreCase("file")) {
			if (scheme.equalsIgnoreCase("file")
					|| !packageHandler.isDirectory(archiveURI)) {
				processor.addTask(installationTaskFactory.getExtractTask());
			} else {
				processor.addTask(installationTaskFactory
						.getCopyArchiveToWorkDirTask());
			}
		} else {
			throw new IllegalArgumentException("URL protocol not supported:"
					+ scheme);
		}

		processor.addTask(installationTaskFactory.getSlCheckPackageTask());

		processor.addTask(installationTaskFactory
				.getCopySLPackageToRepositoryTask());
		processor.addTask(installationTaskFactory.getCreateSLClassloaderTask());
		processor.addTask(installationTaskFactory.getCreateSLStateHolderTask());
		if (scheme.equalsIgnoreCase("file")
				&& !packageHandler.isDirectory(archiveURI)) {
			processor.addTask(installationTaskFactory
					.getCopyArchiveToInstalledDirTask());
		}
		processor.addTask(installationTaskFactory.getSuccessTask());

		return processor;
	}

	/**
	 * Create the uninstallation processor and add all required task for the
	 * complete uninstallation of a shared lib
	 *
	 * @param uninstallationContext
	 *            the uninstallation context
	 * @return the filled uninstallation processor
	 */
	private TaskProcessor createSLUnInstallationProcessor(
			HashMap<String, Object> uninstallationContext) {
		TaskProcessor processor = new TaskProcessor(uninstallationContext, log);

		processor.addTask(installationTaskFactory
				.getComponentShutdownStateCheckingTask());

		processor.addTask(installationTaskFactory.getRemoveSLClassloaderTask());

		processor.addTask(installationTaskFactory.getRemoveSLStateHolderTask());
		processor.addTask(installationTaskFactory.getRemoveSLFolderTask());

		processor.addTask(installationTaskFactory
				.getDeleteSLArchiveFromInstalledDirTask());

		return processor;
	}

	/**
	 * Create the uninstallation processor and add all required task for the
	 * complete uninstallation of a component
	 *
	 * @param uninstallationContext
	 *            the uninstallation context
	 * @return the filled uninstallation processor
	 */
	private TaskProcessor createComponentUnInstallationProcessor(
			HashMap<String, Object> uninstallationContext) {
		TaskProcessor processor = new TaskProcessor(uninstallationContext, log);

		processor.addTask(installationTaskFactory.getDeregisterInstallerTask());

		processor.addTask(installationTaskFactory
				.getComponentStateHolderRemovalTask());
		processor.addTask(installationTaskFactory
				.getComponentFolderRemovalTask());

		processor.addTask(installationTaskFactory
				.getDeleteCompArchiveFromInstalledDirTask());

		return processor;
	}

	/**
	 * Create the component uninstallation context
	 *
	 * @param name
	 *            name of the component to uninstall
	 * @return the Uninstallation Context HashMap
	 */
	private HashMap<String, Object> createComponentUnInstallationContext(
			String name) {
		HashMap<String, Object> uninstallationContext = new HashMap<String, Object>();
		uninstallationContext.put(TaskContextConstants.ENTITY_NAME, name);
		return uninstallationContext;
	}

	/**
	 * Create the shared lib uninstallation context
	 *
	 * @param name
	 *            name of the shared lib to uninstall
	 * @return the Uninstallation Context HashMap
	 */
	private HashMap<String, Object> createSLUnInstallationContext(String name) {
		HashMap<String, Object> uninstallationContext = new HashMap<String, Object>();
		uninstallationContext.put(TaskContextConstants.ENTITY_NAME, name);
		return uninstallationContext;
	}

	/**
	 * @see javax.jbi.management.InstallationServiceMBean#uninstallSharedLibrary(java.lang.String)
	 */
	public boolean uninstallSharedLibrary(String slName) {
		log.start();
		ParameterCheckHelper.isNullParameterWithLog(slName, "slName", log);
		ParameterCheckHelper.isEmptyParameterWithLog(slName, "slName", log);
		boolean result = true;

		String msg;
		HashMap<String, Object> uninstallationContext = null;

		try {

			/*
			 * Create uninstallation execution context
			 */
			uninstallationContext = createSLUnInstallationContext(slName);

			/*
			 * Create uninstallation processor and fill with tasks
			 */
			TaskProcessor processor = createSLUnInstallationProcessor(uninstallationContext);

			/*
			 * Launch the process
			 */
			if (!processor.process()) {
				msg = "An error occured during the uninstallation "
						+ "of the following shared lib: " + slName;
				throw new Exception(msg);
			}
			log.info("Shared Library succesfully uninstalled: "
					+ (String) uninstallationContext
							.get(TaskContextConstants.ENTITY_NAME));

		} catch (Throwable e) {
			msg = "Could not uninstall the shared library: " + slName;
			log.error(msg, e);
			result = false;
		}

		if (result) {
			log.info("Shared library '" + slName + "' correctly uninstalled");
		}
		log.end();

		return result;
	}

	/**
	 * TODO isToBeDeleted parameter isn't used for the moment : this method
	 * completly deletes the component from the container TODO Refactoring
	 * failure cases
	 *
	 * @see javax.jbi.management.InstallationServiceMBean#unloadInstaller(java.lang.String,
	 *      boolean)
	 */
	public boolean unloadInstaller(String componentName,
			boolean isToBeDeleted) {
		log.start();
		ParameterCheckHelper.isNullParameterWithLog(componentName,
				"componentName", log);
		ParameterCheckHelper.isEmptyParameterWithLog(componentName,
				"componentName", log);
		boolean result = true;

		String msg = null;

		HashMap<String, Object> uninstallationContext = null;

		try {

			/*
			 * Create installation execution context
			 */
			uninstallationContext = createComponentUnInstallationContext(componentName);

			/*
			 * Create installation processor and fill with tasks
			 */
			TaskProcessor processor = createComponentUnInstallationProcessor(uninstallationContext);

			/*
			 * Launch the process
			 */
			if (!processor.process()) {
				msg = "An error occured during the uninstallation "
						+ "of the following component: " + componentName;
				throw new Exception(msg);
			}
			log.info("Component succesfully uninstalled: "
					+ (String) uninstallationContext
							.get(TaskContextConstants.ENTITY_NAME));

		} catch (Throwable e) {
			msg = "Component can't be uninstalled: " + componentName;
			log.error(msg, e);
			result = false;
		}
		if (result) {
			log.info("Component '" + componentName + "' correctly uninstalled");
		}
		log.end();
		return result;
	}

	/*
	 * (non-Javadoc)
	 *
	 * @see org.ow2.petals.jbi.management.service.InstallationServiceMBean#getComponentsForInstalledSharedlibrary(java.lang.String)
	 */
	public String[] retrieveComponentsForInstalledSharedlibrary(String slName)
			throws PetalsException {
		throw new PetalsException("Not implemented");
	}

	/*
	 * (non-Javadoc)
	 *
	 * @see org.ow2.petals.jbi.management.service.InstallationServiceMBean#getinstalledSharedLibrary()
	 */
	public String[] retrieveInstalledSharedLibraries() throws PetalsException {
		return systemStateService.getAllInstalledSharedLibNames().toArray(
				new String[0]);
	}

	/**
	 * Initialize the InstallationServiceImpl
	 */
	@LifeCycle(on = LifeCycleType.START)
	protected void start() {
		log = new LoggingUtil(logger);
		log.start();

		/*
		 * Initialize package handler
		 */
		packageHandler = new PackageHandler(logger, repositoryService);

		/*
		 * Initialize installation task factory
		 */
		installationTaskFactory = new InstallationTaskFactory(packageHandler,
				systemStateService, repositoryService, log, logger,
				adminService, router, loaderService, container, autoLoaderService);

		log.end();
	}

	/**
	 * Stop the InstallationServiceImpl
	 */
	@LifeCycle(on = LifeCycleType.STOP)
	public void stop() {
		log.call();
	}

	/**
	 * Shutdown the InstallationServiceMBean. All shared libraries and
	 * components are uninstalled
	 */
	public void shutdown() throws Exception {
		unloadAllComponentInstallers(true);

		uninstallAllSharedLibraries();
	}

	/**
	 * Start all the components
	 *
	 * @return the list of components that have been started
	 */
	public String[] startAllComponents() {
		List<String> startedComponents = new ArrayList<String>();

		for (String name : systemStateService.getAllLoadedComponentNames()) {
			try {
				startComponent(name);
				startedComponents.add(name);
			} catch (JBIException e) {
				log.warning("The following component can't be started: " + name
						+ ". Cause: " + e.getMessage(), e);
			}
		}

		return startedComponents.toArray(new String[startedComponents.size()]);
	}

	/**
	 * Stop all the components
	 *
	 * @return the list of components that have been stopped
	 */
	public String[] stopAllComponents() {
		List<String> stopedComponents = new ArrayList<String>();

		for (String name : systemStateService.getAllLoadedComponentNames()) {
			try {
				stopComponent(name);
				stopedComponents.add(name);
			} catch (JBIException e) {
				log.warning("The following component can't be stoped: " + name
						+ ". Cause: " + e.getMessage(), e);
			}
		}
		return stopedComponents.toArray(new String[stopedComponents.size()]);
	}

	/**
	 * Shutdown all the components
	 *
	 * @return the list of the components that have been shutdowned
	 */
	public String[] shutdownAllComponents() {
		List<String> shutdownComponents = new ArrayList<String>();

		for (String name : systemStateService.getAllLoadedComponentNames()) {
			try {
				shutdownComponent(name);
				shutdownComponents.add(name);
			} catch (JBIException e) {
				log.warning("The following component can't be shutdown: "
						+ name + ". Cause: " + e.getMessage(), e);
			}
		}
		return shutdownComponents
				.toArray(new String[shutdownComponents.size()]);
	}

	/**
	 * Start the component
	 *
	 * @param componentName
	 *            the component to start
	 * @return true if the component has been started
	 * @throws JBIException
	 *             if the component can not be found
	 */
	public boolean startComponent(String componentName) throws JBIException {
		ParameterCheckHelper.isNullParameterWithLog(componentName,
				"componentName", log);
		ParameterCheckHelper.isEmptyParameterWithLog(componentName,
				"componentName", log);

		ComponentLifeCycle componentLifeCycle = adminService
				.getComponentLifeCycleByName(componentName);

		boolean result = false;

		if (componentLifeCycle != null) {
			componentLifeCycle.start();
			result = true;
		} else {
			throw new JBIException("The requested component isn't installed: "
					+ componentName);
		}

		return result;
	}

	/**
	 * Stop a component
	 *
	 * @param componentName
	 *            the component to stop
	 *
	 * @return true if the component has been stopped
	 * @throws JBIException
	 *             if the component can not be found
	 */
	public boolean stopComponent(String componentName) throws JBIException {
		ParameterCheckHelper.isNullParameterWithLog(componentName,
				"componentName", log);
		ParameterCheckHelper.isEmptyParameterWithLog(componentName,
				"componentName", log);

		ComponentLifeCycle componentLifeCycle = adminService
				.getComponentLifeCycleByName(componentName);

		boolean result = false;

		if (componentLifeCycle != null) {
			componentLifeCycle.stop();
			result = true;
		} else {
			throw new JBIException("The requested component isn't installed: "
					+ componentName);
		}

		return result;
	}

	/**
	 * Shutdown a component
	 *
	 * @param componentName
	 *            the component to shutdown
	 * @return true if the component has been shutdowned
	 * @throws JBIException
	 *             if the component has not been found
	 */
	public boolean shutdownComponent(String componentName) throws JBIException {
		ParameterCheckHelper.isNullParameterWithLog(componentName,
				"componentName", log);
		ParameterCheckHelper.isEmptyParameterWithLog(componentName,
				"componentName", log);

		ComponentLifeCycle componentLifeCycle = adminService
				.getComponentLifeCycleByName(componentName);

		boolean result = false;

		if (componentLifeCycle != null) {
			componentLifeCycle.shutDown();
			result = true;
		} else {
			throw new JBIException("The requested component isn't installed: "
					+ componentName);
		}

		return result;
	}

	/**
	 * Uninstall all the components
	 *
	 * @return the list of components that have been stopped
	 */
	public String[] uninstallAllComponents() {
		List<String> uninstalledComponents = new ArrayList<String>();

		for (String name : systemStateService.getAllLoadedComponentNames()) {
			try {
				uninstallComponent(name);
				uninstalledComponents.add(name);
			} catch (JBIException e) {
				log.warning("The following component can't be uninstalled: "
						+ name + ". Cause: " + e.getMessage(), e);
			}
		}
		return uninstalledComponents.toArray(new String[uninstalledComponents
				.size()]);
	}

	/**
	 * Uninstall a component
	 *
	 * @param componentName
	 *            the component to uninstall
	 *
	 * @return true if the component has been uninstalled
	 * @throws JBIException
	 *             if the component can not be found
	 */
	public boolean uninstallComponent(String componentName) throws JBIException {
		ParameterCheckHelper.isNullParameterWithLog(componentName,
				"componentName", log);
		ParameterCheckHelper.isEmptyParameterWithLog(componentName,
				"componentName", log);

		ObjectName instName = loadInstaller(componentName);
		Installer installer = adminService.getInstallerMBeanByName(instName);

		boolean result = false;

		if (installer != null) {
			installer.uninstall();
			result = true;
		} else {
			throw new JBIException("The requested component isn't loaded: "
					+ componentName);
		}

		return result;
	}

	public String[] unloadAllComponentInstallers(boolean force) {

		List<String> removedComponents = new ArrayList<String>();
		for (String compName : systemStateService.getAllLoadedComponentNames()) {
			if (force) {
				if (forceUnloadComponentInstaller(compName)) {
					removedComponents.add(compName);
				}
			} else {
				if (unloadInstaller(compName, true)) {
					removedComponents.add(compName);
				} else {
					log.warning("A component can't be correctly uninstalled: "
							+ compName);
				}
			}
		}

		return removedComponents.toArray(new String[removedComponents.size()]);
	}

	/**
	 * Uninstall all the shared libraries
	 *
	 * @return the list of uninstalled shared libraries
	 */
	public String[] uninstallAllSharedLibraries() {
		List<String> removedSharedLibraries = new ArrayList<String>();
		for (String slName : systemStateService.getAllInstalledSharedLibNames()) {
			if (uninstallSharedLibrary(slName)) {
				removedSharedLibraries.add(slName);
			} else {
				log.warning("A shared library can't be correctly uninstalled: "
						+ slName);
			}
		}
		return removedSharedLibraries.toArray(new String[removedSharedLibraries
				.size()]);
	}

	/**
	 * Force to unload installer of the given component
	 *
	 * @param componentName
	 *            the component to uninstall
	 */
	public boolean forceUnloadComponentInstaller(String componentName) {
		boolean removed = false;

		try {
			tryToShutDownComponent(componentName);
			tryToUninstallComponent(componentName);
			removed = unloadInstaller(componentName, true);
		} catch (JBIException e) {
			log.warning("Force uninstall on the following component failed: "
					+ componentName, e);
		}

		return removed;
	}

	private void tryToUninstallComponent(String componentName)
			throws JBIException {
		ObjectName instName = loadInstaller(componentName);
		Installer installer = adminService.getInstallerMBeanByName(instName);
		if (installer != null) {
			if (installer.isInstalled()) {
				installer.uninstall();
			}
		} else {
			throw new JBIException("The installer of the component '"
					+ componentName + "' is not loaded");
		}
	}

	private void tryToShutDownComponent(String componentName)
			throws JBIException {
		ComponentLifeCycle componentLifeCycle = adminService
				.getComponentLifeCycleByName(componentName);
		if (componentLifeCycle != null) {
			String compState = componentLifeCycle.getCurrentState();
			if (javax.jbi.management.LifeCycleMBean.STARTED.equals(compState)) {
				stopComponent(componentName);
				shutdownComponent(componentName);
			} else if (javax.jbi.management.LifeCycleMBean.STOPPED.equals(compState)) {
				shutdownComponent(componentName);
			}
		}
	}

	public Logger getLogger() {
		return logger;
	}

}
