/**
 * PETALS - PETALS Services Platform.
 * Copyright (c) 2006 EBM Websourcing, http://www.ebmwebsourcing.com/
 *
 * This library is free software; you can redistribute it and/or
 * modify it under the terms of the GNU Lesser General Public
 * License as published by the Free Software Foundation; either
 * version 2.1 of the License, or (at your option) any later version.
 * This library is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 * Lesser General Public License for more details.
 *
 * You should have received a copy of the GNU Lesser General Public
 * License along with this library; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
 *
 * -------------------------------------------------------------------------
 * $Id: SACheckPackageTask.java 154 6 oct. 06 ofabre $
 * -------------------------------------------------------------------------
 */

package org.ow2.petals.jbi.management.task.deployment.deploy;

import java.io.File;
import java.util.ArrayList;
import java.util.Iterator;
import java.util.List;
import java.util.Map;

import javax.xml.namespace.QName;

import org.ow2.petals.PetalsException;
import org.ow2.petals.container.lifecycle.ServiceAssemblyLifeCycle;
import org.ow2.petals.jbi.descriptor.Connection;
import org.ow2.petals.jbi.descriptor.JBIDescriptor;
import org.ow2.petals.jbi.descriptor.ServiceAssembly;
import org.ow2.petals.jbi.descriptor.ServiceUnit;
import org.ow2.petals.jbi.management.admin.AdminServiceMBean;
import org.ow2.petals.jbi.management.task.AbstractCheckPackageTask;
import org.ow2.petals.jbi.management.util.PackageHandler;
import org.ow2.petals.platform.systemstate.SystemStateService;
import org.ow2.petals.util.LoggingUtil;
import org.ow2.petals.util.ParameterCheckHelper;
import org.ow2.petals.util.StringHelper;

/**
 * This task checks the validity of the service assembly package. Some checking
 * steps are described in the specification (section 6.6 of the JSR 208).
 * 
 * @author ofabre - EBM Websourcing
 * 
 */
public class SACheckPackageTask extends AbstractCheckPackageTask {

    protected SystemStateService systemStateService;

    protected AdminServiceMBean adminService;

    public SACheckPackageTask(LoggingUtil log, PackageHandler packageHandler,
            SystemStateService systemStateService, AdminServiceMBean adminService) {
        super(log, packageHandler);
        this.systemStateService = systemStateService;
        this.adminService = adminService;
    }

    @Override
    protected void checkPackage(JBIDescriptor descriptor, File filePackage) throws PetalsException {
        String msg;
        /*
         * Retrieve assembly production element into JBI descriptor
         */
        ServiceAssembly assembly = descriptor.getServiceAssembly();
        if (assembly == null) {
            msg = "Given archive isn't a service assembly archive : "
                    + "missing service assembly production element into jbi descriptor";
            log.error(msg);
            throw new PetalsException(msg);
        }

        String saName = assembly.getIdentification().getName();

        /*
         * Check service assembly identifier
         */
        if (StringHelper.isNullOrEmpty(saName)) {
            msg = "Unique identifier of a service assembly must be non null and non empty ("
                    + filePackage.getAbsolutePath() + ").";
            log.error(msg);
            throw new PetalsException(msg);
        }

        /*
         * Check if sa name is already used
         */
        if (isSADeployed(saName)) {
            msg = "You are trying to deploy a sa with an identifier "
                    + "that is used by a previously successfuly deployed sa (" + saName + ").";
            log.error(msg);
            throw new PetalsException(msg);
        }

        /*
         * Check potential connection elements
         */
        List<Connection> connections = assembly.getConnections();
        if (connections != null) {
            checkConnections(connections);
        }

        /*
         * Check potential service unit elements
         */
        List<ServiceUnit> sus = assembly.getServiceUnits();
        if (sus != null) {
            checkServiceUnits(sus, saName, filePackage);
        }
    }

    private boolean isSADeployed(String saName) {
        return adminService.getServiceAssemblyByName(saName) != null;
    }

    /**
     * Check the validity of connection elements
     * 
     * @param connections
     *            a {@link List} of {@link Connection} element
     * @throws Exception
     *             if consumer interface and consumer endpoint and consumer
     *             service are null or empty at the same time or if (provider
     *             endpoint and provider service) are null or empty
     */
    protected void checkConnections(List<Connection> connections) throws PetalsException {
        String msg = "";

        for (Connection connection : connections) {
            QName consInterface = connection.getConsumerInterfaceName();
            QName consService = connection.getConsumerServiceName();
            String consEndpoint = connection.getConsumerEndpointName();
            QName provService = connection.getProviderServiceName();
            String provEndpoint = connection.getProviderEndpointName();

            if (consInterface == null || StringHelper.isNullOrEmpty(consInterface.getLocalPart())) {
                if (StringHelper.isNullOrEmpty(consEndpoint) || consService == null
                        || StringHelper.isNullOrEmpty(consService.getLocalPart())) {
                    msg = "Consumer interface and (consumer service or consumer endpoint) "
                            + "must not be null or empty at the same time.";
                    log.error(msg);
                    throw new PetalsException(msg);
                }
            }
            if (StringHelper.isNullOrEmpty(provEndpoint)
                    || (provService == null || StringHelper.isNullOrEmpty(provService
                            .getLocalPart()))) {
                msg = "Provider service and provider endpoint) " + "must not be null or empty.";
                log.error(msg);
                throw new PetalsException(msg);
            }
        }
    }

    /**
     * Check the validity of service unit elements
     * 
     * @param sus
     *            a {@link List} of {@link ServiceUnit} element
     * @param saName
     *            the service assembly name
     * @param filePackage
     *            the service assembly package root
     * @throws Exception
     *             if su identifier is null or empty or non unique within the
     *             sa, if target artifact zip and target component are null or
     *             empty, if not all declared su zip are presents in the sa
     *             package or if su has already been deployed to the target
     *             component
     */
    protected void checkServiceUnits(List<ServiceUnit> sus, String saName, File filePackage)
            throws PetalsException {
        String msg = "";

        List<String> suIds = new ArrayList<String>();
        for (ServiceUnit su : sus) {
            String suIdentifier = su.getIdentification().getName();

            // Check su identifier : must be non null, non empty and unique
            // within the sa
            checkSUIdentifier(suIds, suIdentifier, saName);

            suIds.add(suIdentifier);

            // check target artifact zip and target component entries : must
            // be non null and non empty
            String artifactZip = su.getTargetArtifactsZip();
            String targetComponent = su.getTargetComponentName();
            checkTargetComponentAndZip(artifactZip, suIdentifier, targetComponent);

            // check that all declared su zip are presents in the sa package
            checkSUPackageInSAPackage(artifactZip, filePackage);

            // Check if su hasn't already been deployed to the target
            // component
            if (isSUAlreadyDeployedOnTargetedComponent(su)) {
                msg = "You are trying to deploy a su (" + su.getIdentification().getName()
                        + ")that have already been deployed on the targeted component ("
                        + su.getTargetComponentName() + ").";
                log.error(msg);
                throw new PetalsException(msg);
            }

        }
    }

    /**
     * Check su identifier : must be non null, non empty and unique within the
     * sa
     * 
     * @param suIds
     *            the su identifier list within the sa
     * @param suIdentifier
     *            the service unit identifier
     * @param saName
     *            the service assembly name
     * @throws Exception
     *             if su identifier is null, empty and non unique within the sa
     */
    protected void checkSUIdentifier(List<String> suIds, String suIdentifier, String saName)
            throws PetalsException {
        String msg = "";
        if (StringHelper.isNullOrEmpty(suIdentifier)) {
            msg = "Identifier of a service unit must be non null and non empty (SA name : "
                    + saName + ").";
            log.error(msg);
            throw new PetalsException(msg);
        }
        if (suIds.contains(suIdentifier)) {
            msg = "Identifier of a service unit must be unique within a single SA (SA name : "
                    + saName + ").";
            log.error(msg);
            throw new PetalsException(msg);
        }
    }

    /**
     * Check target artifact zip and target component entries : must be non null
     * and non empty
     * 
     * @param artifactZip
     *            the artefact zip path into the sa archive
     * @param suIdentifier
     *            the service unit identifier
     * @param targetComponent
     *            the targeted component name
     * @throws Exception
     *             if target artifact zip or target component entries are null
     *             or empty
     */
    protected void checkTargetComponentAndZip(String artifactZip, String suIdentifier,
            String targetComponent) throws PetalsException {
        String msg = "";
        if (StringHelper.isNullOrEmpty(artifactZip)) {
            msg = "Target artifact zip of a service unit must be non null "
                    + "and non empty (SU name : " + suIdentifier + ").";
            log.error(msg);
            throw new PetalsException(msg);
        }
        if (StringHelper.isNullOrEmpty(targetComponent)) {
            msg = "Target component name of a service unit must be non null "
                    + "and non empty (SU name : " + suIdentifier + ").";
            log.error(msg);
            throw new PetalsException(msg);
        }
    }

    /**
     * Check that all declared su zip are presents in the sa package
     * 
     * @param artifactPath
     *            the artefact path into the sa archive
     * @param filePackage
     *            the service assembly package root
     * @throws Exception
     *             if at least one su is missing into the sa archive
     */
    protected void checkSUPackageInSAPackage(String artifactPath, File filePackage)
            throws PetalsException {
        File artifactFile = new File(filePackage, artifactPath);
        if (!artifactFile.exists()) {
            throw new PetalsException("Can't find artifact (" + artifactPath + ") in SA package: "
                    + filePackage.getAbsolutePath());
        }
    }

    /**
     * Check if the su is already deployed on the targeted component.
     * 
     * @param su
     *            the service unit production element
     * 
     * @return true if the su is already deployed on the targeted component,
     *         false otherwise
     * @throws Exception
     *             if deploymentService.isDeployedServiceUnit fails
     */
    protected boolean isSUAlreadyDeployedOnTargetedComponent(ServiceUnit su) throws PetalsException {
        return isDeployedServiceUnit(su.getTargetComponentName(), su.getIdentification().getName());
    }

    @Override
    protected String extractEntityName(JBIDescriptor descriptor) {
        return descriptor.getServiceAssembly().getIdentification().getName();
    }

    private boolean isDeployedServiceUnit(String componentName, String serviceUnitName) {
        ParameterCheckHelper.isNullParameterWithLog(serviceUnitName, "serviceUnitName", log);
        ParameterCheckHelper.isEmptyParameterWithLog(serviceUnitName, "serviceUnitName", log);
        ParameterCheckHelper.isNullParameterWithLog(componentName, "componentName", log);
        ParameterCheckHelper.isEmptyParameterWithLog(componentName, "componentName", log);
        boolean result = false;
        Map<String, ServiceAssemblyLifeCycle> saLifeCycles = adminService
                .getServiceAssemblyLifeCycles();
        for (Iterator<ServiceAssemblyLifeCycle> iter = saLifeCycles.values().iterator(); iter
                .hasNext()
                && !result;) {
            ServiceAssemblyLifeCycle salc = (ServiceAssemblyLifeCycle) iter.next();
            List<ServiceUnit> sus = salc.getServiceAssembly().getServiceUnits();
            for (Iterator<ServiceUnit> iter2 = sus.iterator(); iter2.hasNext() && !result;) {
                ServiceUnit unit = (ServiceUnit) iter2.next();
                if (componentName.equals(unit.getTargetComponentName())
                        && serviceUnitName.equals(unit.getIdentification().getName())) {
                    result = true;
                }
            }
        }
        return result;
    }

}
