/**
 * PETALS - PETALS Services Platform.
 * Copyright (c) 2005 EBM Websourcing, http://www.ebmwebsourcing.com/
 *
 * This library is free software; you can redistribute it and/or
 * modify it under the terms of the GNU Lesser General Public
 * License as published by the Free Software Foundation; either
 * version 2.1 of the License, or (at your option) any later version.
 * This library is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 * Lesser General Public License for more details.
 *
 * You should have received a copy of the GNU Lesser General Public
 * License along with this library; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
 *
 * -------------------------------------------------------------------------
 * $Id: EndpointRegistry.java 11:27:27 ddesjardins $
 * -------------------------------------------------------------------------
 */

package org.ow2.petals.jbi.messaging.registry;

import static org.ow2.petals.Constants.UNRESOLVED_INTERFACE;

import java.io.IOException;
import java.util.ArrayList;
import java.util.HashMap;
import java.util.List;
import java.util.Map;

import javax.jbi.JBIException;
import javax.jbi.messaging.MessageExchange;
import javax.jbi.servicedesc.ServiceEndpoint;
import javax.management.InstanceNotFoundException;
import javax.management.MBeanException;
import javax.management.MalformedObjectNameException;
import javax.management.ReflectionException;
import javax.xml.namespace.QName;

import org.objectweb.fractal.api.control.IllegalLifeCycleException;
import org.objectweb.fractal.fraclet.annotation.annotations.FractalComponent;
import org.objectweb.fractal.fraclet.annotation.annotations.Interface;
import org.objectweb.fractal.fraclet.annotation.annotations.LifeCycle;
import org.objectweb.fractal.fraclet.annotation.annotations.Monolog;
import org.objectweb.fractal.fraclet.annotation.annotations.Provides;
import org.objectweb.fractal.fraclet.annotation.annotations.Requires;
import org.objectweb.fractal.fraclet.annotation.annotations.type.LifeCycleType;
import org.objectweb.util.monolog.api.Logger;
import org.ow2.petals.PetalsException;
import org.ow2.petals.communication.jmx.DistributedJMXServer;
import org.ow2.petals.communication.jmx.JMXService;
import org.ow2.petals.communication.jndi.client.JNDIService;
import org.ow2.petals.jbi.messaging.endpoint.AbstractEndpoint;
import org.ow2.petals.jbi.messaging.endpoint.ConsumerEndpoint;
import org.ow2.petals.jbi.messaging.endpoint.InternalEndpoint;
import org.ow2.petals.jbi.messaging.endpoint.LinkedEndpoint;
import org.ow2.petals.kernel.configuration.ConfigurationService;
import org.ow2.petals.util.LoggingUtil;
import org.ow2.petals.util.ParameterCheckHelper;
import org.ow2.petals.util.WSDLUtil;
import org.ow2.petals.util.XMLUtil;
import org.w3c.dom.Document;

/**
 * This class is used to manage everything that deals with endpoint in the
 * distributed environment
 *
 * @author ddesjardins - eBMWebsourcing
 */
@FractalComponent
@Provides(interfaces = @Interface(name = "service", signature = org.ow2.petals.jbi.messaging.registry.EndpointRegistry.class))
public class EndpointRegistryImpl implements EndpointRegistry {

    private static final String UNDEFINED = "undefined";

    private static final String KEY_TYPE = "type";

    private static final String KEY_INTERFACE_NAMES = "interfaceNames";

    private static final String KEY_SERVICE_NAME = "serviceName";

    private static final String KEY_ENDPOINT_NAME = "endpointName";

    private static final String KEY_CONTAINER_NAME = "containerName";

    private static final String KEY_COMPONENT_NAME = "componentName";

    private static final String TYPE_INTERNAL = "internal";

    private static final String TYPE_EXTERNAL = "external";

    /**
     * Logger wrapper
     */
    protected LoggingUtil log;

    /**
     * Monolog Logger instance.
     */
    @Monolog(name = "logger")
    protected Logger logger;

    /**
     * Registry
     */
    protected JNDIRegistry registry;

    /**
     * Manager of the distributed JMX server
     */
    @Requires(name = "jmx", signature = org.ow2.petals.communication.jmx.JMXService.class)
    protected JMXService serverManager;

    /**
	 * JNDI service fractal component
	 */
	@Requires(name = "jndi", signature = org.ow2.petals.communication.jndi.client.JNDIService.class)
	protected JNDIService jndiService;

	/**
	 * Configuration service fractal component
	 */
	@Requires(name = "configuration", signature = org.ow2.petals.kernel.configuration.ConfigurationService.class)
	protected ConfigurationService configurationService;

    /**
     * @see EndpointRegistry#activateEndpoint(QName, String, ConsumerEndpoint)
     */
    public InternalEndpoint activateEndpoint(QName serviceName, String endpointName,
            ConsumerEndpoint address) throws JBIException {
        log.start();
        if (serviceName == null) {
            throw new IllegalArgumentException("Service name must be non null");
        }
        if (endpointName == null) {
            throw new IllegalArgumentException("Endpoint name must be non null");
        }
        if (address == null) {
            throw new IllegalArgumentException("Consumer endpoint must be non null");
        }
        InternalEndpoint se = new InternalEndpoint(serviceName, endpointName, address
                .getComponentName(), address.getContainerName(),address.getSubDomainName(), this);
        registry.registerInternalEndpoint(se);
        log.end();
        return se;
    }

    /**
     * @see EndpointRegistry#createConnection(QName, QName, String)
     */
    public void createConnection(QName consInterface, QName provService, String provEndpoint)
            throws JBIException {
        log.start();
        if (consInterface == null) {
            throw new IllegalArgumentException("Interface name must be non null");
        }
        if (provService == null) {
            throw new IllegalArgumentException("Provider service name must be non null");
        }
        if (provEndpoint == null) {
            throw new IllegalArgumentException("Provider endpoint name must be non null");
        }
        LinkedEndpoint linkedEndpoint = new LinkedEndpoint(consInterface, provService,
                provEndpoint, this);
        registry.registerConnection(linkedEndpoint);
        log.end();
    }

    /**
     * @see EndpointRegistry#createConnection(QName, String, QName, String)
     */
    public void createConnection(QName consService, String consEndpoint, QName provService,
            String provEndpoint) throws JBIException {
        log.start();
        if (consService == null) {
            throw new IllegalArgumentException("Consumer service name must be non null");
        }
        if (consEndpoint == null) {
            throw new IllegalArgumentException("Consumer endpoint name must be non null");
        }
        if (provService == null) {
            throw new IllegalArgumentException("Provider service name must be non null");
        }
        if (provEndpoint == null) {
            throw new IllegalArgumentException("Provider endpoint name must be non null");
        }
        LinkedEndpoint linkedEndpoint = new LinkedEndpoint(consService, consEndpoint, provService,
                provEndpoint, this);
        registry.registerConnection(linkedEndpoint);
        log.end();
    }

    /**
     * @see EndpointRegistry#deactivateEndpoint(ServiceEndpoint)
     */
    public void deactivateEndpoint(ServiceEndpoint endpoint) throws JBIException {
        log.start();
        if (endpoint == null) {
            throw new IllegalArgumentException("Endpoint must be non null");
        }
        registry.unregisterInternalEndpoint((AbstractEndpoint) endpoint);
        log.end();
    }

    /**
     * @see EndpointRegistry#deleteConnection(QName, QName, String)
     */
    public void deleteConnection(QName consInterface, QName provService, String provEndpoint)
            throws JBIException {
        log.start();
        if (consInterface == null) {
            throw new IllegalArgumentException("Consumer interface name must be non null");
        }
        if (provService == null) {
            throw new IllegalArgumentException("Provider service name must be non null");
        }
        if (provEndpoint == null) {
            throw new IllegalArgumentException("Provider endpoint name must be non null");
        }
        LinkedEndpoint linkedEndpoint = new LinkedEndpoint(consInterface, provService,
                provEndpoint, this);
        registry.unregisterConnection(linkedEndpoint);
        log.end();
    }

    /**
     * @see EndpointRegistry#deleteConnection(QName, String, QName, String)
     */
    public void deleteConnection(QName consService, String consEndpoint, QName provService,
            String provEndpoint) throws JBIException {
        log.start();
        if (consService == null) {
            throw new IllegalArgumentException("Consumer service name must be non null");
        }
        if (consEndpoint == null) {
            throw new IllegalArgumentException("Consumer endpoint name must be non null");
        }
        if (provService == null) {
            throw new IllegalArgumentException("Provider service name must be non null");
        }
        if (provEndpoint == null) {
            throw new IllegalArgumentException("Provider endpoint name must be non null");
        }
        LinkedEndpoint linkedEndpoint = new LinkedEndpoint(consService, consEndpoint, provService,
                provEndpoint, this);
        registry.unregisterConnection(linkedEndpoint);
        log.end();
    }

    /**
     * @see EndpointRegistry#deregisterExternalEndpoint(ServiceEndpoint)
     */
    public void deregisterExternalEndpoint(ServiceEndpoint endpoint) throws JBIException {
        log.start();
        if (endpoint == null) {
            throw new IllegalArgumentException("Endpoint must be non null");
        }
        registry.unregisterExternalEndpoint(endpoint);
        log.end();
    }

    /**
     * @see EndpointRegistry#getEndpoint(QName, String)
     *
     * <ol>
     * <li>Retreive the named endpoint</li>
     * <li>If found, update the endpoint service of the found endpoint with
     * this</li>
     * </ol>
     */
    public ServiceEndpoint getEndpoint(QName service, String name) {
        log.call();
        return getEndpoint(service, name, true);
    }

    /**
     * @see EndpointRegistry#getEndpoint(QName, String,boolean)
     *
     * <ol>
     * <li>Retrieve the named endpoint</li>
     * <li>If found, update the endpoint service of the found endpoint with
     * this</li>
     * </ol>
     */
    public ServiceEndpoint getEndpoint(QName service, String name, boolean resolveLink) {
        log.start();
        if (service == null) {
            throw new IllegalArgumentException("Service name must be non null");
        }
        if (name == null) {
            throw new IllegalArgumentException("Endpoint name must be non null");
        }
        AbstractEndpoint abstractEndpoint = null;
        try {
            abstractEndpoint = registry.getInternalEndpoint(service, name, resolveLink);
        } catch (RegistryException e) {
            // no endpoint found, return null;
        }
        if (abstractEndpoint != null && abstractEndpoint.getEndpointRegistry() == null) {
            abstractEndpoint.setEndpointRegistry(this);
        }
        log.end();
        return abstractEndpoint;
    }

    /**
     * @see EndpointRegistry#getEndpointDescriptorForEndpoint(ServiceEndpoint)
     */
    public Document getEndpointDescriptorForEndpoint(ServiceEndpoint endpoint) throws JBIException {
        log.start();
        if (endpoint == null) {
            throw new IllegalArgumentException("Endpoint must be non null");
        }
        Document document = null;
        try {

            // Update the new endpoints
            updateNewEndpoints();

            // Recover the endpoint
            AbstractEndpoint abstractEndpoint = registry.getInternalEndpoint(endpoint
                    .getServiceName(), endpoint.getEndpointName());

            document = abstractEndpoint.getEndpointDescriptor();
        } catch (RegistryException e) {
            log.error("Error while validating endpoint " + endpoint.getEndpointName(), e);
        }
        log.end();
        return document;
    }

    /**
     * @see EndpointRegistry#getExternalEndpointsForInterface(QName)
     */
    public ServiceEndpoint[] getExternalEndpointsForInterface(QName interfaceName) {
        log.start();
        ServiceEndpoint[] endpoints = new ServiceEndpoint[0];
        try {
            endpoints = registry.getExternalEndpointsForInterface(interfaceName);
            // Update the EndpointRegistry field of the found endpoints
            for (int i = 0; i < endpoints.length; i++) {
                AbstractEndpoint abstractEndpoint = (AbstractEndpoint) endpoints[i];
                if (abstractEndpoint.getEndpointRegistry() == null) {
                    abstractEndpoint.setEndpointRegistry(this);
                    endpoints[i] = abstractEndpoint;
                }
            }
        } catch (RegistryException e) {
            log.error("Error while getting external endpoints for interface " + interfaceName, e);
        }
        log.end();
        return endpoints;
    }

    /**
     * @see EndpointRegistry#getExternalEndpointsForService(QName)
     */
    public ServiceEndpoint[] getExternalEndpointsForService(QName serviceName) {
        log.start();
        if (serviceName == null) {
            throw new IllegalArgumentException("Service name must be non null");
        }
        ServiceEndpoint[] endpoints = null;
        endpoints = registry.getExternalEndpointsForService(serviceName);
        // Update the EndpointRegistry field of the found endpoints
        for (int i = 0; i < endpoints.length; i++) {
            AbstractEndpoint abstractEndpoint = (AbstractEndpoint) endpoints[i];
            if (abstractEndpoint.getEndpointRegistry() == null) {
                abstractEndpoint.setEndpointRegistry(this);
                endpoints[i] = abstractEndpoint;
            }
        }
        log.end();
        return endpoints;
    }

    /**
     * Return the interface for the specified endpoint
     *
     * @param endpoint
     *            endpoint on which we want the interface names; must be non
     *            null
     * @return array of interfaces QName
     */
    public QName[] getInterfacesForEndpoint(AbstractEndpoint endpoint) {
        log.start();
        if (endpoint == null) {
            throw new IllegalArgumentException("Endpoint must be non null");
        }
        QName[] interfs = new QName[0];
        AbstractEndpoint abstractEndpoint = null;
        try {
            updateNewEndpoints();
            abstractEndpoint = registry.getInternalEndpoint(endpoint.getServiceName(), endpoint
                    .getEndpointName());
            interfs = abstractEndpoint.getInterfaces();
        } catch (Exception e) {
            log.error("Error while getting interfaces for endpoint " + endpoint.getEndpointName(),
                    e);
        }
        log.end();
        return interfs;
    }

    /**
     * Update all the new endpoints and remove them
     *
     * @throws JBIException
     *             if a problem occurs
     */
    protected void updateNewEndpoints() throws JBIException {

        for (AbstractEndpoint abstractEndpoint : registry.retrieveNewEndpoints()) {
            AbstractEndpoint endpoint = null;
            try {
                endpoint = updateEndpointDescriptionAndInterfaces(abstractEndpoint);
                registry.validateEndpoint(endpoint);
            } catch (JBIException e) {
                log.error("Problem while getting description and interfaces for the endpoint "
                        + abstractEndpoint.getEndpointName(), e);
            }
        }
        // We clean the new endpoints context
        registry.cleanNewEndpoints();

    }

    /**
     * @see EndpointRegistry#getInternalEndpointsForInterface(QName)
     */
    public ServiceEndpoint[] getInternalEndpointsForInterface(QName interfaceName) {
        log.start();
        ServiceEndpoint[] endpoints = new ServiceEndpoint[0];
        try {
            // Update the new endpoints
            updateNewEndpoints();

            endpoints = registry.getInternalEndpointsForInterface(interfaceName);
            for (int i = 0; i < endpoints.length; i++) {
                AbstractEndpoint abstractEndpoint = (AbstractEndpoint) endpoints[i];
                if (abstractEndpoint.getEndpointRegistry() == null) {
                    abstractEndpoint.setEndpointRegistry(this);
                    endpoints[i] = abstractEndpoint;
                }
            }
        } catch (RegistryException e) {
            log.error("Error while getting internal endpoints for interface " + interfaceName, e);
        } catch (JBIException e) {
            log.error("Error while getting internal endpoints for interface " + interfaceName, e);
        }
        log.end();
        return endpoints;
    }

    /**
     * @see EndpointRegistry#getInternalEndpointsForService(QName)
     */
    public ServiceEndpoint[] getInternalEndpointsForService(QName serviceName) {
        log.start();
        if (serviceName == null) {
            throw new IllegalArgumentException("Service name must be non null");
        }
        ServiceEndpoint[] endpoints = new ServiceEndpoint[0];
        endpoints = registry.getInternalEndpointsForService(serviceName);
        for (int i = 0; i < endpoints.length; i++) {
            AbstractEndpoint abstractEndpoint = (AbstractEndpoint) endpoints[i];
            if (abstractEndpoint.getEndpointRegistry() == null) {
                abstractEndpoint.setEndpointRegistry(this);
                endpoints[i] = abstractEndpoint;
            }
        }
        log.end();
        return endpoints;
    }

    /**
     * @see EndpointRegistry#isExchangeWithConsumerOkayForComponent(InternalEndpoint,
     *      MessageExchange)
     */
    public boolean isExchangeWithConsumerOkayForComponent(InternalEndpoint internalEndpoint,
            MessageExchange exchange) {
        log.start();
        ParameterCheckHelper.isNullParameterWithLog(internalEndpoint, "Endpoint", log);
        boolean test = true;
        try {
            // Retrieve a connection to the JMX server of the container were the
            // component is deployed
            DistributedJMXServer server = serverManager.createDistributedJMXServer(internalEndpoint
                    .getContainerName());
            // Ask the component if it is ok to accept the exchange
            Boolean bool = (Boolean) server.invoke(server.getAdminServiceMBeanName(),
                    DistributedJMXServer.IS_OK_WITH_CONS, new Object[] {
                            internalEndpoint.getComponentName(), internalEndpoint, exchange },
                    new String[] { String.class.getName(), ServiceEndpoint.class.getName(),
                            MessageExchange.class.getName() });
            test = bool.booleanValue();
        } catch (Exception e) {
            log.error("Problem in isExchangeWithConsumerOkayForComponent", e);
        }
        log.end();
        return test;
    }

    /**
     * @see EndpointRegistry#isExchangeWithProviderOkayForComponent(InternalEndpoint,
     *      MessageExchange)
     */
    public boolean isExchangeWithProviderOkayForComponent(InternalEndpoint internalEndpoint,
            MessageExchange exchange) {
        log.start();
        if (internalEndpoint == null) {
            throw new IllegalArgumentException("Endpoint must be non null");
        }
        boolean test = true;
        try {
            // Retrieve a connection to the JMX server of the container were the
            // component is deployed
            DistributedJMXServer server = serverManager.createDistributedJMXServer(internalEndpoint
                    .getContainerName());
            // Retreive the endpoint description
            Boolean bool = (Boolean) server.invoke(server.getAdminServiceMBeanName(),
                    DistributedJMXServer.IS_OK_WITH_PROV, new Object[] {
                            internalEndpoint.getComponentName(), internalEndpoint, exchange },
                    new String[] { String.class.getName(), ServiceEndpoint.class.getName(),
                            MessageExchange.class.getName() });
            test = bool.booleanValue();
        } catch (Exception e) {
            log.error("Problem in isExchangeWithProviderOkayForComponent", e);
        }
        log.end();
        return test;
    }

    /**
     * @see EndpointRegistry#registerExternalEndpoint(ServiceEndpoint)
     */
    public void registerExternalEndpoint(ServiceEndpoint externalEndpoint) throws JBIException {
        log.start();
        if (externalEndpoint == null) {
            throw new IllegalArgumentException("Endpoint must be non null");
        }
        registry.registerExternalEndpoint(externalEndpoint);
        log.end();
    }

    @LifeCycle(on = LifeCycleType.START)
    public void start() throws IllegalLifeCycleException {
        log = new LoggingUtil(logger);
        log.call();
        registry = new JNDIRegistry(configurationService, jndiService,log);
		try {
			registry.initContexts();
		} catch (RegistryException e) {
			throw new IllegalLifeCycleException(e.toString());
		}

    }

    @LifeCycle(on = LifeCycleType.STOP)
    public void stop() throws IllegalLifeCycleException {
    	registry = null;
        log.call();
    }

    /**
     * This method will update the endpoint description and its interfaces then
     * return it
     *
     * @param endpoint
     *            endpoint to update
     * @return updated endpoint
     * @throws JBIException
     */
    private AbstractEndpoint updateEndpointDescriptionAndInterfaces(AbstractEndpoint endpoint)
            throws JBIException {
        log.start();

        String description = null;
        try {
         // Get a connection to the JMX server of the container were the
            // endpoint is deployed
            DistributedJMXServer server = serverManager.createDistributedJMXServer(endpoint.getContainerName());
            // Retrieve the endpoint description
            description = (String) server.invoke(server.getAdminServiceMBeanName(),
                    DistributedJMXServer.GET_SERVICE_DESCRITION, new Object[] {
                            endpoint.getComponentName(), endpoint }, new String[] {
                            "java.lang.String", "javax.jbi.servicedesc.ServiceEndpoint" });
        } catch (PetalsException e) {
            throw new JBIException("Failed to retrieve the interfaces from the component '"
                    + endpoint.getComponentName() + "' for the endpoint '"
                    + endpoint.getEndpointName(), e);
        } catch (InstanceNotFoundException e) {
            throw new JBIException("Failed to retrieve the interfaces from the component '"
                    + endpoint.getComponentName() + "' for the endpoint '"
                    + endpoint.getEndpointName(), e);
        } catch (MalformedObjectNameException e) {
            throw new JBIException("Failed to retrieve the interfaces from the component '"
                    + endpoint.getComponentName() + "' for the endpoint '"
                    + endpoint.getEndpointName(), e);
        } catch (MBeanException e) {
            throw new JBIException("Failed to retrieve the interfaces from the component '"
                    + endpoint.getComponentName() + "' for the endpoint '"
                    + endpoint.getEndpointName(), e);
        } catch (ReflectionException e) {
            throw new JBIException("Failed to retrieve the interfaces from the component '"
                    + endpoint.getComponentName() + "' for the endpoint '"
                    + endpoint.getEndpointName(), e);
        } catch (NullPointerException e) {
            throw new JBIException("Failed to retrieve the interfaces from the component '"
                    + endpoint.getComponentName() + "' for the endpoint '"
                    + endpoint.getEndpointName(), e);
        } catch (IOException e) {
            throw new JBIException("Failed to retrieve the interfaces from the component '"
                    + endpoint.getComponentName() + "' for the endpoint '"
                    + endpoint.getEndpointName(), e);
        }

        List<QName> interfaces = new ArrayList<QName>();
        // Get the whole description
        endpoint.setDescription(description);

        // Get the interfaces
        try {
            interfaces = WSDLUtil.getInterfacesForService(XMLUtil
                    .createDocumentFromString(description), endpoint.getServiceName());
        } catch (PetalsException e) {
            QName unresolved = UNRESOLVED_INTERFACE;
            log.warning("No retrieved interfaces for the endpoint '" + endpoint.getEndpointName()
                    + "'. Set its interface to " + unresolved);
            interfaces.add(unresolved);
        }

        endpoint.setInterfaces(interfaces.toArray(new QName[0]));

        log.end();
        return endpoint;
    }

    public Document getDescription(String serviceName, String endpointName) throws Exception {
        ServiceEndpoint serviceEndpoint = getEndpoint(QName.valueOf(serviceName), endpointName);
        try {
            return getEndpointDescriptorForEndpoint(serviceEndpoint);
        } catch (JBIException e) {
            String msg = "Failed to retrieve endpoint description for : servicename=" + serviceName
                    + " and endpointname=" + endpointName;
            log.error(msg, e);
            throw new Exception(msg);
        }
    }

    public List<Map<String, Object>> getAllExternalEndpoints() {
        List<Map<String, Object>> externalEndpointsMap = new ArrayList<Map<String, Object>>();
        ServiceEndpoint[] externalEndpoints = getExternalEndpointsForInterface(null);
        for (ServiceEndpoint serviceEndpoint : externalEndpoints) {
            AbstractEndpoint abstractEndpoint = (AbstractEndpoint) serviceEndpoint;
            Map<String, Object> endpointMap = new HashMap<String, Object>();
            endpointMap.put(KEY_COMPONENT_NAME, abstractEndpoint.getComponentName());
            endpointMap.put(KEY_CONTAINER_NAME, abstractEndpoint.getContainerName());
            endpointMap.put(KEY_ENDPOINT_NAME, abstractEndpoint.getEndpointName());
            endpointMap.put(KEY_SERVICE_NAME, abstractEndpoint.getServiceName());
            List<String> interfaceNames = interfaceQNameToString(abstractEndpoint);
            endpointMap.put(KEY_INTERFACE_NAMES, interfaceNames.toArray(new String[0]));
            endpointMap.put(KEY_TYPE, TYPE_EXTERNAL);
            externalEndpointsMap.add(endpointMap);
        }
        return externalEndpointsMap;
    }

    private List<String> interfaceQNameToString(AbstractEndpoint abstractEndpoint) {
        QName[] interfaces = abstractEndpoint.getInterfaces();
        List<String> interfaceNames = new ArrayList<String>();
        if (interfaces != null && interfaces.length > 0) {
            for (QName interfaceName : interfaces) {
                interfaceNames.add(interfaceName.toString());
            }
        } else {
            interfaceNames.add(UNDEFINED);
        }
        return interfaceNames;
    }

    public List<Map<String, Object>> getAllInternalEndpoints() {
        List<Map<String, Object>> internalEndpointsMap = new ArrayList<Map<String, Object>>();
        ServiceEndpoint[] internalEndpoints = getInternalEndpointsForInterface(null);
        for (ServiceEndpoint serviceEndpoint : internalEndpoints) {
            AbstractEndpoint abstractEndpoint = (AbstractEndpoint) serviceEndpoint;
            Map<String, Object> endpointMap = new HashMap<String, Object>();
            endpointMap.put(KEY_COMPONENT_NAME, abstractEndpoint.getComponentName());
            endpointMap.put(KEY_CONTAINER_NAME, abstractEndpoint.getContainerName());
            endpointMap.put(KEY_ENDPOINT_NAME, abstractEndpoint.getEndpointName());
            endpointMap.put(KEY_SERVICE_NAME, abstractEndpoint.getServiceName().toString());
            List<String> interfaceNames = interfaceQNameToString(abstractEndpoint);
            endpointMap.put(KEY_INTERFACE_NAMES, interfaceNames.toArray(new String[0]));
            if (serviceEndpoint instanceof LinkedEndpoint) {
                LinkedEndpoint linkEP = (LinkedEndpoint) serviceEndpoint;
                endpointMap.put(KEY_TYPE, linkEP.getToServiceName().toString() + "@"
                        + linkEP.getToEndpointName());
            } else {
                endpointMap.put(KEY_TYPE, TYPE_INTERNAL);
            }
            internalEndpointsMap.add(endpointMap);
        }
        return internalEndpointsMap;
    }

    public List<Map<String, Object>> getAllEndpoints() {
        List<Map<String, Object>> allEndpointsMap = getAllExternalEndpoints();
        allEndpointsMap.addAll(getAllInternalEndpoints());
        return allEndpointsMap;
    }

}
