/**
 * PETALS - PETALS Services Platform.
 * Copyright (c) 2007 EBM Websourcing, http://www.ebmwebsourcing.com/
 *
 * This library is free software; you can redistribute it and/or
 * modify it under the terms of the GNU Lesser General Public
 * License as published by the Free Software Foundation; either
 * version 2.1 of the License, or (at your option) any later version.
 * This library is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 * Lesser General Public License for more details.
 *
 * You should have received a copy of the GNU Lesser General Public
 * License along with this library; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
 *
 * -------------------------------------------------------------------------
 * $Id$
 * -------------------------------------------------------------------------
 */

package org.ow2.petals.jbi.messaging.routing.EndpointChooser;

import static org.ow2.petals.jbi.messaging.routing.EndpointChooser.Strategy.DEFAULT_STRATEGY;
import static org.ow2.petals.jbi.messaging.routing.EndpointChooser.Strategy.HIGHEST_STRATEGY;
import static org.ow2.petals.jbi.messaging.routing.EndpointChooser.Strategy.RANDOM_STRATEGY;
import static org.ow2.petals.jbi.messaging.routing.EndpointChooser.Strategy.STANDALONE_STRATEGY;
import static org.ow2.petals.jbi.messaging.routing.EndpointChooser.Strategy.PROPERTY_ROUTING_STRATEGY;

import java.util.ArrayList;
import java.util.HashMap;
import java.util.Hashtable;
import java.util.List;
import java.util.Map;
import java.util.StringTokenizer;

import javax.jbi.messaging.MessageExchange;
import javax.jbi.servicedesc.ServiceEndpoint;

import org.ow2.petals.PetalsException;
import org.ow2.petals.communication.topology.TopologyService;
import org.ow2.petals.jbi.messaging.endpoint.InternalEndpoint;
import org.ow2.petals.jbi.messaging.routing.RoutingException;
import org.ow2.petals.kernel.configuration.ConfigurationService;

/**
 * The enpoint chooser strategy.
 * 
 * @author nsalatge - eBM WebSourcing
 * @author Roland NAUDIN - eBM WebSourcing
 * 
 */
public final class EndpointChooserStrategy {

    /**
     * The strategy separator.
     */
    public static final String STRATEGY_SEPARATOR = ",";

    /**
     * Map of Strategy implementations. only one implementation per strategy
     * TODO: add a pool of implementation to bring better performance
     * (multi-thread context)
     */
    private Map<String, Strategy> strategiesMap;

    /**
     * The Topology service
     */
    private TopologyService topologyService;

    /**
     * The Configuration service
     */
    private ConfigurationService configurationService;

    /**
     * Default constructor.
     * 
     */
    public EndpointChooserStrategy(ConfigurationService configurationService,
            TopologyService topologyService) {
        this.topologyService = topologyService;
        this.configurationService = configurationService;
        this.strategiesMap = new Hashtable<String, Strategy>();
    }

    /**
     * Elects the address in the list of endpoints as the message recipient,
     * with the defined strategy. Ask the consumer and the provider if they are
     * Okay for the exchange. TODO: remove the synchronize and use a pool of
     * Strategy instances instead
     * 
     * @param endpoints
     *            the list of endpoints to select an endpoint from
     * @param exchange
     *            the exchange
     * @return the service endpoint selected or null
     * @throws RoutingException
     *             impossible to realize this strategy
     * @throws PetalsException
     * @throws PetalsException
     */
    public synchronized ServiceEndpoint choose(final List<ServiceEndpoint> endpoints,
            final MessageExchange exchange) throws RoutingException {

        Strategy strategy = this.analyseAndGetStrategy(exchange);

        ServiceEndpoint result = strategy.electEndpoint(endpoints, exchange);

        return result;

    }

    /**
     * Get the weighting for each provider.
     * 
     * @param endpoints
     *            the endpoints
     * @param localPonderation
     *            the local weighting
     * @param remoteActiveWeighting
     *            the remote active weighting
     * @param remoteInactiveWeighting
     *            the remote inactive weighting
     * @return all weightings
     * @throws RoutingException
     *             impossible to contact the container topology
     */
    public Map<ServiceEndpoint, Float> getWeightingEndpoints(final List<ServiceEndpoint> endpoints,
            final float localWeighting, final float remoteActiveWeighting,
            final float remoteInactiveWeighting) throws RoutingException {

        Map<ServiceEndpoint, Float> result = new HashMap<ServiceEndpoint, Float>();
        try {
            for (ServiceEndpoint provider : endpoints) {
                if (((InternalEndpoint) provider).getContainerName().equals(
                        configurationService.getContainerConfiguration().getName())) {
                    // Local endpoint
                    result.put(provider, new Float(localWeighting));
                } else

                if (topologyService.isContainerStarted(((InternalEndpoint) provider)
                        .getContainerName())) {
                    // Active remote endpoint
                    result.put(provider, new Float(remoteActiveWeighting));
                } else {
                    // Inactive Remote endpoint
                    result.put(provider, new Float(remoteInactiveWeighting));
                }
            }
        } catch (PetalsException e) {
            throw new RoutingException(e);
        }
        return result;
    }

    /**
     * Get the strategy implementation.
     * 
     * @param type
     *            the strategy type
     * @throws RoutingException
     *             Impossible to find strategy implementation
     */
    private Strategy getStrategy(final String type) throws RoutingException {
        Strategy strategy = strategiesMap.get(type);
        if (strategy == null) {
            if (STANDALONE_STRATEGY.equals(type)) {
                strategy = strategiesMap.put(STANDALONE_STRATEGY, new StandaloneStrategy());
            } else if (RANDOM_STRATEGY.equals(type)) {
                strategy = strategiesMap.put(RANDOM_STRATEGY, new RandomStrategy(this));
            } else if (HIGHEST_STRATEGY.equals(type)) {
                strategy = strategiesMap.put(HIGHEST_STRATEGY, new HighestStrategy(this));
            }
        } else {
            throw new RoutingException("The " + type + " routing strategy is unknown");
        }

        return strategy;
    }

    /**
     * Analyze and get the strategy.
     * 
     * @param exchange
     *            the message exchange. Must be non-null
     * @throws RoutingException
     *             impossible to retrieve the endpoint chooser strategy
     * @return the Strategy implementation matching the parameters
     */
    private Strategy analyseAndGetStrategy(final MessageExchange exchange) throws RoutingException {
        Strategy strategy = null;

        if (exchange == null) {
            throw new NullPointerException("The exchange must be non null");
        }

        String strategyParametersSerialized = (String) exchange
                .getProperty(PROPERTY_ROUTING_STRATEGY);
        if (strategyParametersSerialized == null) {
            strategy = this.getDefaultRoutingStrategy();
        } else {
            List<String> strategyParameters = new ArrayList<String>();
            StringTokenizer st = new StringTokenizer(strategyParametersSerialized,
                    STRATEGY_SEPARATOR);
            while (st.hasMoreTokens()) {
                strategyParameters.add(st.nextToken().trim());
            }

            if (strategyParameters.size() == 0) {
                throw new RoutingException(
                        "The first parameter of the routing strategy is mandatory");
            } else {
                strategy = this.getStrategy(strategyParameters.get(0));
                strategyParameters.remove(0);
                strategy.validateStrategyParameters(strategyParameters);
            }

        }

        return strategy;
    }

    /**
     * Get the analyzed default routing strategy.
     * 
     * @return the analyzed default routing strategy.
     * @throws RoutingException
     *             impossible to analyze the default routing strategy.
     */
    private Strategy getDefaultRoutingStrategy() throws RoutingException {

        if (!this.strategiesMap.containsKey(DEFAULT_STRATEGY)) {
            String strategyParametersSerialized = configurationService.getContainerConfiguration()
                    .getRoutingStrategy();
            StringTokenizer st = new StringTokenizer(strategyParametersSerialized,
                    EndpointChooserStrategy.STRATEGY_SEPARATOR);

            List<String> strategyParameters = new ArrayList<String>();
            while (st.hasMoreTokens()) {
                strategyParameters.add(st.nextToken().trim());
            }

            if (strategyParameters.size() == 0) {
                throw new RoutingException(
                        "The first parameter of the routing strategy is mandatory");
            }

            Strategy defaultStrategy = null;
            if (STANDALONE_STRATEGY.equals(strategyParameters.get(0))) {
                defaultStrategy = new StandaloneStrategy();
            } else if (RANDOM_STRATEGY.equals(strategyParameters.get(0))) {
                defaultStrategy = new RandomStrategy(this);
            } else if (HIGHEST_STRATEGY.equals(strategyParameters.get(0))) {
                defaultStrategy = new HighestStrategy(this);
            } else {
                throw new RoutingException("The " + strategyParameters.get(0)
                        + " routing strategy is unknown");
            }

            strategyParameters.remove(0);
            defaultStrategy.validateStrategyParameters(strategyParameters);

            this.strategiesMap.put(DEFAULT_STRATEGY, defaultStrategy);
        }

        return this.strategiesMap.get(DEFAULT_STRATEGY);
    }

}
