/**
 * PETALS - PETALS Services Platform.
 * Copyright (c) 2007 EBM Websourcing, http://www.ebmwebsourcing.com/
 *
 * This library is free software; you can redistribute it and/or
 * modify it under the terms of the GNU Lesser General Public
 * License as published by the Free Software Foundation; either
 * version 2.1 of the License, or (at your option) any later version.
 * This library is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 * Lesser General Public License for more details.
 *
 * You should have received a copy of the GNU Lesser General Public
 * License along with this library; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
 *
 * -------------------------------------------------------------------------
 * $Id$
 * -------------------------------------------------------------------------
 */

package org.ow2.petals.kernel.configuration;

import java.io.File;
import java.io.FileInputStream;
import java.io.IOException;
import java.net.URL;
import java.util.Collection;
import java.util.HashSet;
import java.util.Hashtable;
import java.util.Map;
import java.util.Properties;

import javax.xml.bind.JAXBException;

import org.objectweb.fractal.fraclet.annotation.annotations.FractalComponent;
import org.objectweb.fractal.fraclet.annotation.annotations.Interface;
import org.objectweb.fractal.fraclet.annotation.annotations.LifeCycle;
import org.objectweb.fractal.fraclet.annotation.annotations.Monolog;
import org.objectweb.fractal.fraclet.annotation.annotations.Provides;
import org.objectweb.fractal.fraclet.annotation.annotations.type.LifeCycleType;
import org.objectweb.util.monolog.api.Logger;
import org.ow2.petals.jbi.messaging.routing.Router;
import org.ow2.petals.kernel.configuration.ContainerConfiguration.ContainerState;
import org.ow2.petals.topology.TopologyBuilder;
import org.ow2.petals.topology.TopologyException;
import org.ow2.petals.topology.TopologyHelper;
import org.ow2.petals.topology.generated.Container;
import org.ow2.petals.topology.generated.DomainMode;
import org.ow2.petals.topology.generated.Subdomain;
import org.ow2.petals.topology.generated.Topology;
import org.ow2.petals.util.LoggingUtil;
import org.ow2.petals.util.StringHelper;

/**
 * Class implementing the {@link ConfigurationService} interface.
 * 
 * @author Roland NAUDIN - EBMWebsourcing
 * @since Petals 2.0
 */

@FractalComponent
@Provides(interfaces = @Interface(name = "service", signature = org.ow2.petals.kernel.configuration.ConfigurationService.class))
public class ConfigurationServiceImpl implements ConfigurationService {

    /**
     * Logging utilities
     */
    protected LoggingUtil log;

    @Monolog(name = "logger")
    protected Logger logger;

    private ContainerConfiguration containerConfiguration;

    private DomainConfiguration domainConfiguration;

    private Properties serverProperties;

    private Topology topology;

    /*
     * (non-Javadoc)
     * 
     * @see org.ow2.petals.kernel.configuration.ConfigurationService#getContainerConfiguration()
     */
    public ContainerConfiguration getContainerConfiguration() {
        log.call();
        return this.containerConfiguration;
    }

    /*
     * (non-Javadoc)
     * 
     * @see org.ow2.petals.kernel.configuration.ConfigurationService#getDomainConfiguration()
     */
    public DomainConfiguration getDomainConfiguration() {
        log.call();
        return this.domainConfiguration;
    }

    /*
     * (non-Javadoc)
     * 
     * @see org.ow2.petals.kernel.configuration.ConfigurationService#getStaticTopology()
     */
    public Topology getStaticTopology() throws ConfigurationException {
        log.call();
        if (this.domainConfiguration.getMode().equals(DomainConfiguration.DomainMode.DYNAMIC)) {
            throw new ConfigurationException("The domain is not in static mode");
        }
        return topology;
    }

    /*
     * (non-Javadoc)
     * 
     * @see org.ow2.petals.kernel.configuration.ConfigurationService#setContainerState(org.ow2.petals.kernel.configuration.ContainerConfiguration.ContainerState)
     */
    public void setContainerState(ContainerState state) {
        this.containerConfiguration.setState(state);
    }

    /*
     * (non-Javadoc)
     * 
     * @see org.ow2.petals.kernel.configuration.ConfigurationService#verifyContainerTopology(org.ow2.petals.kernel.configuration.ContainerConfiguration,
     *      org.ow2.petals.kernel.configuration.DomainConfiguration)
     */
    public void verifyContainerTopology(ContainerConfiguration newContainerConfiguration,
            DomainConfiguration newDomainConfiguration) throws TopologyException {
        log.start();

        verifyDomain(newDomainConfiguration);

        verifyContainer(newContainerConfiguration);

        log.end();
    }

    /**
     * Start the fractal component.
     * 
     * @throws ConfigurationException
     * 
     * @throws JAXBException
     *             The parsing of the Topology file has failed
     * @throws IOException
     *             A configuration resource cannot be reached or read
     * @throws TopologyException
     */
    @LifeCycle(on = LifeCycleType.START)
    protected void start() throws ConfigurationException {
        log = new LoggingUtil(logger);
        log.start();

        try {
            // retrieve resources
            URL serverPropsURL = this.getClass().getResource(SERVER_PROPS_FILE);
            if (serverPropsURL == null) {

                throw new IOException("Failed to reach the resource '" + SERVER_PROPS_FILE + "'");
            }
            File serverPropsFile = new File(serverPropsURL.getPath());
            URL topologyURL = this.getClass().getResource(TOPOLOGY_FILE);
            if (topologyURL == null) {
                throw new IOException("Failed to reach the resource '" + TOPOLOGY_FILE + "'");
            }
            File topologyFile = new File(topologyURL.getPath());

            containerConfiguration = new ContainerConfiguration();
            containerConfiguration.init();
            domainConfiguration = new DomainConfiguration();

            createConfiguration(topologyFile, serverPropsFile);

        } catch (IOException e) {
            // log the error here as Fractal layer is removing the Exception
            // stack
            log.error("Failed to start Configuration service", e);
            throw new ConfigurationException(e);
        } catch (ConfigurationException e) {
            // log the error here as Fractal layer is removing the Exception
            // stack
            log.error("Failed to start Configuration service", e);
            throw e;
        } catch (JAXBException e) {
            // log the error here as Fractal layer is removing the Exception
            // stack
            log.error("Failed to start Configuration service", e);
            throw new ConfigurationException(e);
        } catch (TopologyException e) {
            // log the error here as Fractal layer is removing the Exception
            // stack
            log.error("Failed to start Configuration service", e);
            throw new ConfigurationException(e);
        } catch (NullPointerException e) {
            // log the error here as Fractal layer is removing the Exception
            // stack
            log.error("Failed to start Configuration service", e);
            throw new ConfigurationException(e);
        }

        log.end();
    }

    @LifeCycle(on = LifeCycleType.STOP)
    protected void stop() throws Exception {
        log.call();
    }

    /**
     * Create the container configuration based on the topology and server
     * files. Basic consistency is controlled between the 2 configuration files.
     * The {@code serverPropsFile} and <{@code topologyFile} fields must be non
     * null.
     * 
     * @param topologyFile
     *            The topology.xml file
     * @param serverPropsFile
     *            The server.properties file
     * 
     * @throws JAXBException
     * @throws TopologyException
     * @throws IOException
     * @throws ConfigurationException
     *             Configuration consistency error
     */
    private void createConfiguration(File topologyFile, File serverPropsFile) throws JAXBException,
            TopologyException, IOException, ConfigurationException {
        log.start();

        // build the topology object
        TopologyBuilder topologyBuilder = TopologyBuilder.getInstance();
        topology = topologyBuilder.createTopology(topologyFile);

        // retrieve the server properties
        serverProperties = new Properties();
        serverProperties.load(new FileInputStream(serverPropsFile));

        String containerName = serverProperties
                .getProperty(ConfigurationService.PROPERTY_CONTAINER_NAME);
        if (containerName == null) {
            containerName = System.getProperty(PROPERTY_CONTAINER_NAME);
        }

        containerName = validateTopology(containerName);

        fillGlobalTopology();

        fillLocalTopology(containerName);

        fillGlobalProperties();

        // in standalone mode the container is always started
        if (org.ow2.petals.kernel.configuration.DomainConfiguration.DomainMode.STANDALONE
                .equals(domainConfiguration.getMode())) {
            containerConfiguration.setState(ContainerState.STARTED);
        }

        log.debug("Configuration used :");
        log.debug(containerConfiguration);

        log.end();
    }

    /**
     * Validate the consistency of the topology. The {@code topology} and the
     * {@code containerConfiguration} fields must be non {@code null}.
     * 
     * @param containerName
     *            The name of the local container retrieved from the
     *            environment, can be {@ null}
     * @throws ConfigurationException
     */
    private String validateTopology(String containerName) throws ConfigurationException {
        log.start();

        if (topology.getDomain().getSubDomain().size() == 0) {
            throw new ConfigurationException(
                    "A domain must contain at least one subdomain definition");
        }

        for (Subdomain subdomain : topology.getDomain().getSubDomain()) {
            if (subdomain.getContainer().size() == 0) {
                throw new ConfigurationException(
                        "A subdomain must contain at least one container definition");
            }
        }

        if (!DomainMode.STATIC.equals(topology.getDomain().getMode())) {
            if (topology.getDomain().getSubDomain().size() > 1
                    || topology.getDomain().getSubDomain().get(0).getContainer().size() > 1)
                throw new ConfigurationException(
                        "Standalone or Dynamic mode accepts only one sub-domain and one container definition.");
        }

        // validate domain consistency
        if (DomainMode.STATIC.equals(topology.getDomain().getMode())) {
            if (containerName == null) {
                throw new ConfigurationException(
                        "The property 'petals.container.name' must be set in static distributed mode");
            }
            if (TopologyHelper.findContainer(containerName, topology) == null) {
                throw new ConfigurationException("Cannot find the definition for the container "
                        + containerName + " in the topology definition.");
            }
            if (DomainMode.STATIC.equals(topology.getDomain().getMode())) {
                validateStaticTopology();
            }
        } else {
            containerName = topology.getDomain().getSubDomain().get(0).getContainer().get(0)
                    .getName();
        }

        // validate JNDI consistency
        if (topology.getDomain().getJndi() != null) {
            if ("".equals(topology.getDomain().getJndi().getFactory().trim())) {
                throw new ConfigurationException(
                        "JNDI factory classname must be set to reference an external JNDI server");
            }
            if (topology.getDomain().getJndi().getHost() == null
                    || topology.getDomain().getJndi().getPort() <= 0) {
                throw new ConfigurationException(
                        "JNDI host and port must be set to reference an external JNDI server with factory : '"
                                + topology.getDomain().getJndi().getFactory() + "'");
            }
        }

        // validate network consistency
        if (topology.getDomain().getMulticast() == null
                && DomainMode.DYNAMIC.equals(topology.getDomain().getMode())) {
            throw new ConfigurationException("Multicast domain is mandatory in dynamic mode");
        }

        // validate containers configuration against the domain mode
        if (!DomainMode.STANDALONE.equals(topology.getDomain().getMode())) {
            Collection<Container> containers = TopologyHelper.findAllContainers(topology);
            for (Container container : containers) {
                if (container.getConfiguration().getNetworkservice() == null) {
                    throw new ConfigurationException(
                            "Network service must be defined in the container '"
                                    + container.getName() + "' configuration");
                }
                if (container.getConfiguration().getJoramservice() == null) {
                    throw new ConfigurationException(
                            "Joram service must be defined in the container '"
                                    + container.getName() + "' configuration");
                }
                if (container.getConfiguration().getDreamservice() == null) {
                    throw new ConfigurationException(
                            "Dream service must be defined in the container '"
                                    + container.getName() + "' configuration");
                }
            }
        }

        log.end();
        return containerName;
    }

    /**
     * Validate the consistency of the static topology
     * 
     * @throws ConfigurationException
     */
    private void validateStaticTopology() throws ConfigurationException {
        log.start();

        // Check all the sub domains names
        Collection<String> subdomainNameList = new HashSet<String>();
        for (Subdomain subdomain : topology.getDomain().getSubDomain()) {
            if (!subdomainNameList.add(subdomain.getName())) {
                throw new ConfigurationException("Duplicate sub-domain name '"
                        + subdomain.getName() + "'");
            }
        }

        // Check all the containers
        Collection<String> nameList = new HashSet<String>();
        Map<String, HashSet<Integer>> portMap = new Hashtable<String, HashSet<Integer>>();
        for (Container container : TopologyHelper.findAllContainers(topology)) {
            if (container.getHost() == null) {
                throw new ConfigurationException("The container named '" + container.getName()
                        + "' must declare its host element in static distributed mode");
            }

            HashSet<Integer> portList;
            if (!portMap.containsKey(container.getHost())) {
                portList = new HashSet<Integer>();
                portMap.put(container.getHost(), portList);
            } else {
                portList = portMap.get(container.getHost());
            }
            if (!portList.add(container.getConfiguration().getNetworkservice().getPort())) {
                throw new ConfigurationException("Duplicate port "
                        + container.getConfiguration().getNetworkservice().getPort() + " on host '"
                        + container.getHost() + "'");
            }
            if (container.getConfiguration().getJmxservice().getRmiport() != 0) {
                if (!portList.add(container.getConfiguration().getJmxservice().getRmiport())) {
                    throw new ConfigurationException("Duplicate port "
                            + container.getConfiguration().getJmxservice().getRmiport()
                            + " on host '" + container.getHost() + "'");
                }
            }
            if (!portList.add(container.getConfiguration().getJoramservice().getDomainPort())) {
                throw new ConfigurationException("Duplicate port "
                        + container.getConfiguration().getJoramservice().getDomainPort()
                        + " on host '" + container.getHost() + "'");
            }
            if (!portList.add(container.getConfiguration().getJoramservice().getTcpPort())) {
                throw new ConfigurationException("Duplicate port "
                        + container.getConfiguration().getJoramservice().getTcpPort()
                        + " on host '" + container.getHost() + "'");
            }
            if (!portList.add(container.getConfiguration().getDreamservice().getPort())) {
                throw new ConfigurationException("Duplicate port "
                        + container.getConfiguration().getDreamservice().getPort() + " on host '"
                        + container.getHost() + "'");
            }
            if (!nameList.add(container.getName())) {
                throw new ConfigurationException("Duplicate container name '" + container.getName()
                        + "'");
            }
        }

        log.end();
    }

    /**
     * Fill the global topology.
     */
    private void fillGlobalTopology() {
        log.start();

        domainConfiguration.setDomain(topology.getDomain().getName());
        if (DomainMode.STANDALONE.equals(topology.getDomain().getMode())) {
            domainConfiguration
                    .setMode(org.ow2.petals.kernel.configuration.DomainConfiguration.DomainMode.STANDALONE);
            // set the default value of the routing strategy of standalone mode
            containerConfiguration.setRoutingStrategy(Router.DEFAULT_STANDALONE_STRATEGY);
        } else if (DomainMode.STATIC.equals(topology.getDomain().getMode())) {
            domainConfiguration
                    .setMode(org.ow2.petals.kernel.configuration.DomainConfiguration.DomainMode.STATIC);
        } else {
            domainConfiguration
                    .setMode(org.ow2.petals.kernel.configuration.DomainConfiguration.DomainMode.DYNAMIC);
        }

        if (topology.getDomain().getMulticast() != null) {
            domainConfiguration.setNetworkMuticastIp(topology.getDomain().getMulticast().getIp());
            domainConfiguration.setNetworkMuticastPort(topology.getDomain().getMulticast()
                    .getPort());
        } else if (DomainMode.STATIC.equals(topology.getDomain().getMode())) {
            // fill the map of unicast network addresses
            Collection<Container> topologyCollection = TopologyHelper.findAllContainers(topology);
            Map<String, String> networkUnicastAddressMap = new Hashtable<String, String>();
            for (Container container : topologyCollection) {
                networkUnicastAddressMap.put(container.getName(), container.getHost() + "["
                        + container.getConfiguration().getNetworkservice().getPort() + "]");
            }
            domainConfiguration.setNetworkUnicastAddressMap(networkUnicastAddressMap);
        }

        if (topology.getDomain().getJndi() != null) {
            domainConfiguration.setJndiFactory(topology.getDomain().getJndi().getFactory());
            domainConfiguration.setJndiHost(topology.getDomain().getJndi().getHost());
            domainConfiguration.setJndiPort(topology.getDomain().getJndi().getPort());
        }

        log.end();
    }

    /**
     * Fill the container configuration with its topology part. The
     * {@code containerConfiguration} and {@code topology} fields must be non
     * {@code null}.
     * 
     * @param container
     *            name The local container name
     * @throws ConfigurationException
     *             Configuration consistency error
     */
    private void fillLocalTopology(String containerName) throws ConfigurationException {
        log.start();

        Container localContainer = TopologyHelper.findContainer(containerName, topology);

        if (localContainer == null) {
            throw new ConfigurationException(
                    "Failed to retrieve the container configuration for the container named '"
                            + containerName + "'");
        }

        containerConfiguration.setName(containerName);

        containerConfiguration.setSubdomain(TopologyHelper.findSubdomain(localContainer, topology)
                .getName());
        
        containerConfiguration.setUser(localContainer.getConfiguration().getUser());
        containerConfiguration.setPassword(localContainer.getConfiguration().getPassword());
        
        if (localContainer.getHost() != null) {
            containerConfiguration.setHost(localContainer.getHost());
        }

        // Network service, can be null
        if (localContainer.getConfiguration().getNetworkservice() != null) {
            containerConfiguration.setNetworkBindingPort(localContainer.getConfiguration()
                    .getNetworkservice().getPort());
        }

        // JMX Service
        containerConfiguration.setJmxRMIConnectorPort(localContainer.getConfiguration()
                .getJmxservice().getRmiport());
        containerConfiguration.setJmxJNDIConnectorPort(localContainer.getConfiguration()
                .getJmxservice().getJndiport());
        

        // JORAM Service
        if (localContainer.getConfiguration().getJoramservice() != null) {
            containerConfiguration.setJoramDomainPort(localContainer.getConfiguration()
                    .getJoramservice().getDomainPort());
            containerConfiguration.setJoramTCPPort(localContainer.getConfiguration().getJoramservice()
                    .getTcpPort());
        }

        // DREAM Service
        if (localContainer.getConfiguration().getDreamservice() != null) {
            containerConfiguration.setDreamTCPPort(localContainer.getConfiguration()
                    .getDreamservice().getPort());
        }

        log.end();
    }

    /**
     * Fill the container configuration with its global properties part. The
     * {@code containerConfiguration} and {@code serverProperties} fields must
     * be non {@code null}.
     */
    private void fillGlobalProperties() {
        log.start();

        for (Object key : serverProperties.keySet()) {
            if (key instanceof String) {
                if (((String) key).equals(PROPERTY_REPOSITORY_PATH)) {
                    containerConfiguration.setRepositoryDirectoryPath(serverProperties
                            .getProperty((String) key));
                } else if (((String) key).equals(PROPERTY_EXCHANGE_VALIDATION)) {
                    containerConfiguration.setExchangeValidation(new Boolean(serverProperties
                            .getProperty((String) key)));
                } else if (((String) key).equals(PROPERTY_TRANSPORT_QOS)) {
                    containerConfiguration.setTransportQOS(serverProperties
                            .getProperty((String) key));
                } else if (((String) key).equals(PROPERTY_ROUTING_STRATEGY)) {
                    containerConfiguration.setRoutingStrategy(serverProperties
                            .getProperty((String) key));
                } else if (((String) key).equals(PROPERTY_TASK_TIMEOUT)) {
                    containerConfiguration.setTaskTimeout(new Long(serverProperties
                            .getProperty((String) key)));
                }
            }
        }

        log.end();
    }

    /**
     * Verify the given new domain configuration against the topology.
     * 
     * @param newDomainConfiguration
     * @throws TopologyException
     */
    private void verifyDomain(DomainConfiguration newDomainConfiguration) throws TopologyException {
        log.debug("Check the domain configuration of the new Container...");

        if (!newDomainConfiguration.getMode().equals(domainConfiguration.getMode())) {
            throw new TopologyException("Domain mode mismatch");
        }

        // JNDI service
        if (!StringHelper.equal(newDomainConfiguration.getJndiFactory(), domainConfiguration
                .getJndiFactory())) {
            throw new TopologyException("JNDI factory mismatch");
        }
        if (!StringHelper.equal(newDomainConfiguration.getJndiHost(), domainConfiguration
                .getJndiHost())) {
            throw new TopologyException("JNDI Host mismatch");
        }
        if (newDomainConfiguration.getJndiPort() != domainConfiguration.getJndiPort()) {
            throw new TopologyException("JNDI Port mismatch");
        }
        log.debug("Domain configuration of the new Container is OK.");
    }

    /**
     * Verify the given new container configuration against the topology.
     * 
     * @param newContainerConfiguration
     * @throws TopologyException
     */
    private void verifyContainer(ContainerConfiguration newContainerConfiguration)
            throws TopologyException {

        log.debug("Check the container configuration of the new Container.");
        if (DomainMode.STATIC.equals(topology.getDomain().getMode())) {
            Container newContainer = TopologyHelper.findContainer(newContainerConfiguration
                    .getName(), topology);

            if (newContainer == null) {
                throw new TopologyException("Container name unknown");
            }
            log.debug("... check container " + newContainer.getName() + " ...");
            if (!StringHelper.equal(newContainer.getConfiguration().getUser(),
                    newContainerConfiguration.getUser())) {
                throw new TopologyException("User mismatch");
            }
            if (!StringHelper.equal(newContainer.getConfiguration().getPassword(),
                    newContainerConfiguration.getPassword())) {
                throw new TopologyException("Password mismatch");
            }

            // network Service
            if (newContainer.getConfiguration().getNetworkservice().getPort() != newContainerConfiguration
                    .getNetworkBindingPort()) {
                throw new TopologyException("Network binding port mismatch");
            }
            // JMX Service
            if (newContainer.getConfiguration().getJmxservice().getRmiport() != newContainerConfiguration
                    .getJmxRMIConnectorPort()) {
                throw new TopologyException("JMX RMI connector port mismatch");
            }
            // JORAM service
            if (newContainer.getConfiguration().getJoramservice().getDomainPort() != newContainerConfiguration
                    .getJoramDomainPort()) {
                throw new TopologyException("JORAM Domain port mismatch");
            }
            if (newContainer.getConfiguration().getJoramservice().getTcpPort() != newContainerConfiguration
                    .getJoramTCPPort()) {
                throw new TopologyException("JORAM TCP port mismatch");
            }
            if (newContainer.getConfiguration().getDreamservice().getPort() != newContainerConfiguration
                    .getDreamTCPPort()) {
                throw new TopologyException("DREAM TCP port mismatch");
            }
        } else if (newContainerConfiguration.getHost().equals(containerConfiguration.getHost())) {
            if (newContainerConfiguration.getName() == containerConfiguration.getName()) {
                throw new TopologyException("Duplicate container name '"
                        + containerConfiguration.getName() + "'");
            }

        }
        log.debug("Container configuration of the new Container is OK.");
    }
}
