/**
 * PETALS - PETALS Services Platform.
 * Copyright (c) 2007 EBM Websourcing, http://www.ebmwebsourcing.com/
 *
 * This library is free software; you can redistribute it and/or
 * modify it under the terms of the GNU Lesser General Public
 * License as published by the Free Software Foundation; either
 * version 2.1 of the License, or (at your option) any later version.
 * This library is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 * Lesser General Public License for more details.
 *
 * You should have received a copy of the GNU Lesser General Public
 * License along with this library; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
 *
 * -------------------------------------------------------------------------
 * $Id$
 * -------------------------------------------------------------------------
 */

package org.ow2.petals.kernel.configuration;

import java.io.File;
import java.io.IOException;
import java.io.Serializable;

import org.ow2.petals.jbi.messaging.routing.Router;
import org.ow2.petals.transport.Transporter;
import org.ow2.petals.platform.repository.RepositoryService;
import org.ow2.petals.util.SystemUtil;

/**
 * This class holds the overall configuration of a PEtALS container. It stores
 * both the local topology information and the global server properties
 * information.<br>
 * The configuration service retrieves the configuration when it is started and
 * creates an instance of this class. The instance is exposed on the distributed
 * JNDI server.
 * 
 * @author alouis, Roland NAUDIN - eBMWebSourcing
 * @since Petals 2.0
 * 
 */
public final class ContainerConfiguration implements Serializable {

    /**
     * the "unique" ID for serialization
     */
    private static final long serialVersionUID = 6543213857L;

    /**
     * Container state enumeration.
     */
    public static enum ContainerState {
        /**
         * Started state
         */
        STARTED,
        /**
         * Stopped state
         */
        STOPPED,
        /**
         * unknown state
         */
        UNKNOWN
    };

    // ********** Global ************//

    /**
     * Subdomain .
     */
    private String subdomain;

    /**
     * Container name.
     */
    private String name;

    /**
     * Host of the petals container.
     */
    private String host;

    /**
     * State of the container.
     */
    private ContainerState state;

    /**
     * Install directory.
     */
    private String installDirectoryPath;

    /**
     * Repository directory path.
     */
    private String repositoryDirectoryPath;

    /**
     * Exchange validation.<br>
     * If <code>true</code>, used by the Address resolver to know if a
     * component is agree to receive a exchange from a consumer.
     */
    private boolean exchangeValidation;

    /**
     * Transport Quality of Service. {@see Transporter}
     */
    private String transportQOS = Transporter.DEFAULT_POLICY;

    /**
     * Routing Strategy. {@see EndpointChooserStrategy}
     */
    private String routingStrategy = Router.DEFAULT_PLATFORM_STRATEGY;

    /**
     * The timeout for processing the installation and deployment tasks.
     */
    private long taskTimeout;

    /**
     * Login to connect to the node
     */
    private String user;

    /**
     * Password to connect to the node
     */
    private String password;

    // ********** Network ************//

    /**
     * Network port. Port used for PEtALS unicast messages
     */
    private int networkBindingPort;

    // ************ JMX **************//

    /**
     * Port of the RMI adaptor for JMX
     */
    private int jmxRMIConnectorPort;

    /**
     * Port of the JNDI adaptor for JMX
     */
    private int jmxJNDIConnectorPort;

    // ***** JOrAM Transporter *****//

    /**
     * Port of the Joram domain
     */
    private int joramDomainPort;

    /**
     * Port of the Joram TCP communication
     */
    private int joramTCPPort;

    // ***** DReAM Transporter *****//

    /**
     * Port of the Dream TCP communication
     */
    private int dreamTCPPort;

    public ContainerConfiguration() {
        super();
    }

    public void init() throws ConfigurationException {
        this.state = ContainerState.UNKNOWN;
        this.host = SystemUtil.resolveLocalIP();
        try {
            this.installDirectoryPath = SystemUtil.getPetalsInstallDirectory().getCanonicalPath();
        } catch (IOException e) {
            throw new ConfigurationException("The PEtALS install directory is not valid : "
                    + SystemUtil.getPetalsInstallDirectory().getAbsolutePath());
        }
        this.repositoryDirectoryPath = this.installDirectoryPath + File.separator
                + RepositoryService.DEFAULT_REPOSITORY_PATH;

    }

    public String getSubdomain() {
        return subdomain;
    }

    public void setSubdomain(String subdomain) {
        this.subdomain = subdomain;
    }

    public String getName() {
        return name;
    }

    public String getHost() {
        return host;
    }

    public ContainerState getState() {
        return state;
    }

    public String getInstallDirectoryPath() {
        return installDirectoryPath;
    }

    public String getRepositoryDirectoryPath() {
        return repositoryDirectoryPath;
    }

    public boolean isExchangeValidation() {
        return exchangeValidation;
    }

    public String getTransportQOS() {
        return transportQOS;
    }

    public String getRoutingStrategy() {
        return routingStrategy;
    }

    public long getTaskTimeout() {
        return taskTimeout;
    }

    public int getNetworkBindingPort() {
        return networkBindingPort;
    }

    public String getUser() {
        return user;
    }

    public String getPassword() {
        return password;
    }

    public int getJmxRMIConnectorPort() {
        return jmxRMIConnectorPort;
    }

    public int getJmxJNDIConnectorPort() {
        return jmxJNDIConnectorPort;
    }

    public int getJoramDomainPort() {
        return joramDomainPort;
    }

    public int getJoramTCPPort() {
        return joramTCPPort;
    }

    public int getDreamTCPPort() {
        return dreamTCPPort;
    }

    protected void setName(String name) {
        this.name = name;
    }

    protected void setHost(String host) {
        this.host = host;
    }

    public void setState(ContainerState state) {
        this.state = state;
    }

    protected void setInstallDirectoryPath(String installDirectoryPath) {
        this.installDirectoryPath = installDirectoryPath;
    }

    protected void setRepositoryDirectoryPath(String repositoryDirectoryPath) {
        this.repositoryDirectoryPath = repositoryDirectoryPath;
    }

    protected void setExchangeValidation(boolean exchangeValidation) {
        this.exchangeValidation = exchangeValidation;
    }

    protected void setTransportQOS(String transportQOS) {
        this.transportQOS = transportQOS;
    }

    protected void setRoutingStrategy(String routingStrategy) {
        this.routingStrategy = routingStrategy;
    }

    protected void setTaskTimeout(long taskTimeout) {
        this.taskTimeout = taskTimeout;
    }

    protected void setNetworkBindingPort(int networkBindingPort) {
        this.networkBindingPort = networkBindingPort;
    }

    protected void setUser(String user) {
        this.user = user;
    }

    protected void setPassword(String password) {
        this.password = password;
    }

    protected void setJmxRMIConnectorPort(int jmxRMIConnectorPort) {
        this.jmxRMIConnectorPort = jmxRMIConnectorPort;
    }

    protected void setJmxJNDIConnectorPort(int jmxJNDIConnectorPort) {
        this.jmxJNDIConnectorPort = jmxJNDIConnectorPort;
    }

    protected void setJoramDomainPort(int joramDomainPort) {
        this.joramDomainPort = joramDomainPort;
    }

    protected void setJoramTCPPort(int joramTCPPort) {
        this.joramTCPPort = joramTCPPort;
    }

    protected void setDreamTCPPort(int dreamTCPPort) {
        this.dreamTCPPort = dreamTCPPort;
    }

    @Override
    public String toString() {
        StringBuilder toStringBuilder = new StringBuilder();
        toStringBuilder.append("Container " + this.name + " on " + this.host + " in sub-domain "
                + subdomain + "\n  Configuration :\n  ---------------");
        if (installDirectoryPath != null) {
            toStringBuilder.append("\n  - InstallDirectory:" + this.installDirectoryPath);
        }
        if (repositoryDirectoryPath != null) {
            if (installDirectoryPath != null) {
                toStringBuilder.append("\n  - ");
            }
            toStringBuilder.append("RepositoryDirectory:" + this.repositoryDirectoryPath);
        }
        if (installDirectoryPath != null || repositoryDirectoryPath != null) {
            toStringBuilder.append("\n  - ");
        }
        toStringBuilder.append("ExchangeValidation:" + this.exchangeValidation);
        toStringBuilder.append("\n  - TransportQOS:" + this.transportQOS);
        toStringBuilder.append("\n  - RoutingStrategy:" + this.routingStrategy);
        toStringBuilder.append("\n  - TaskTimeout:" + this.taskTimeout);
        toStringBuilder.append("\n  - User:" + ((this.user != null) ? this.user : "none"));
        toStringBuilder.append("\n  - Password:"
                + ((this.password != null) ? this.password : "none"));
        toStringBuilder.append("\n  - NetworkBindingPort:"
                + ((this.networkBindingPort != 0) ? this.networkBindingPort : "none"));
        toStringBuilder.append("\n  - JMXRMIConnectionPort:" + this.jmxRMIConnectorPort);
        toStringBuilder.append("\n  - JMXJNDIConnectionPort:" + this.jmxJNDIConnectorPort);
        toStringBuilder.append("\n  - JoramDomainPort:"
                + ((this.joramDomainPort != 0) ? this.joramDomainPort : "none"));
        toStringBuilder.append("\n  - JoramTCPPort:"
                + ((this.joramTCPPort != 0) ? this.joramTCPPort : "none"));
        toStringBuilder.append("\n  - DreamTCPPort:"
                + ((this.dreamTCPPort != 0) ? this.dreamTCPPort : "none"));
        toStringBuilder.append("\n  - State:" + this.state.name());

        return toStringBuilder.toString();
    }

    /**
     * Duplicate the given container configuration in an exported form.
     * 
     * @param containerConfiguration
     *            The original container configuration
     * @return the exported form
     */
    public static ContainerConfiguration getExportedForm(
            ContainerConfiguration containerConfiguration) {
        ContainerConfiguration exportedForm = new ContainerConfiguration();

        exportedForm.name = containerConfiguration.name;
        exportedForm.subdomain = containerConfiguration.subdomain;
        exportedForm.host = containerConfiguration.host;
        exportedForm.exchangeValidation = containerConfiguration.exchangeValidation;
        exportedForm.transportQOS = containerConfiguration.transportQOS;
        exportedForm.routingStrategy = containerConfiguration.routingStrategy;
        exportedForm.taskTimeout = containerConfiguration.taskTimeout;
        exportedForm.networkBindingPort = containerConfiguration.networkBindingPort;
        exportedForm.user = containerConfiguration.user;
        exportedForm.password = containerConfiguration.password;
        exportedForm.jmxRMIConnectorPort = containerConfiguration.jmxRMIConnectorPort;
        exportedForm.jmxJNDIConnectorPort = containerConfiguration.jmxJNDIConnectorPort;
        exportedForm.joramDomainPort = containerConfiguration.joramDomainPort;
        exportedForm.joramTCPPort = containerConfiguration.joramTCPPort;
        exportedForm.dreamTCPPort = containerConfiguration.dreamTCPPort;
        exportedForm.state = containerConfiguration.state;
        return exportedForm;
    }

    @Override
    public int hashCode() {
        final int prime = 31;
        int result = 1;
        result = prime * result + subdomain.hashCode();
        result = prime * result + dreamTCPPort;
        result = prime * result + (exchangeValidation ? 1231 : 1237);
        result = prime * result + ((host == null) ? 0 : host.hashCode());
        result = prime * result + ((user == null) ? 0 : user.hashCode());
        result = prime * result + ((password == null) ? 0 : password.hashCode());
        result = prime * result + jmxRMIConnectorPort;
        result = prime * result + jmxJNDIConnectorPort;
        result = prime * result + joramDomainPort;
        result = prime * result + joramTCPPort;
        result = prime * result + ((name == null) ? 0 : name.hashCode());
        result = prime * result + networkBindingPort;
        result = prime * result + ((routingStrategy == null) ? 0 : routingStrategy.hashCode());
        result = prime * result + (int) (taskTimeout ^ (taskTimeout >>> 32));
        result = prime * result + ((transportQOS == null) ? 0 : transportQOS.hashCode());
        return result;
    }

    @Override
    public boolean equals(Object obj) {
        if (this == obj) {
            return true;
        }
        if (obj == null) {
            return false;
        }
        boolean result = false;
        if (obj instanceof ContainerConfiguration) {
            ContainerConfiguration othersContainerConfiguration = (ContainerConfiguration) obj;
            result = (othersContainerConfiguration.name.equals(this.name)
                    && othersContainerConfiguration.host.equals(this.host)
                    && othersContainerConfiguration.exchangeValidation == this.exchangeValidation
                    && othersContainerConfiguration.transportQOS.equals(this.transportQOS)
                    && othersContainerConfiguration.routingStrategy.equals(this.routingStrategy)
                    && othersContainerConfiguration.taskTimeout == this.taskTimeout
                    && othersContainerConfiguration.networkBindingPort == this.networkBindingPort
                    && othersContainerConfiguration.user.equals(this.user)
                    && othersContainerConfiguration.password.equals(this.password)
                    && othersContainerConfiguration.jmxRMIConnectorPort == this.jmxRMIConnectorPort
                    && othersContainerConfiguration.jmxJNDIConnectorPort == this.jmxJNDIConnectorPort
                    && othersContainerConfiguration.joramDomainPort == this.joramDomainPort
                    && othersContainerConfiguration.joramTCPPort == this.joramTCPPort && othersContainerConfiguration.dreamTCPPort == this.dreamTCPPort);
        }
        return result;
    }

}
