/**
 * PETALS: PETALS Services Platform
 * Copyright (C) 2005 EBM WebSourcing
 *
 * This library is free software; you can redistribute it and/or
 * modify it under the terms of the GNU Lesser General Public
 * License as published by the Free Software Foundation; either
 * version 2.1 of the License, or any later version.
 *
 * This library is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 * Lesser General Public License for more details.
 *
 * You should have received a copy of the GNU Lesser General Public
 * License along with this library; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307
 * USA.
 *
 * Initial developer(s): EBM WebSourcing
 * --------------------------------------------------------------------------
 * $Id: Transporter.java,v 0.3 2005/07/22 10:24:27 alouis Exp $
 * --------------------------------------------------------------------------
 */

package org.ow2.petals.transport;

import java.util.List;

import org.ow2.petals.PetalsLifeCycle;
import org.ow2.petals.jbi.messaging.exchange.MessageExchangeImpl;

/**
 * The <code>Transporter</code> is used to send a <code>MessageExchange</code>
 * from a Petals container to another one. <br>
 * The local <code>Router</code> calls
 * <code>send(exchange,distantContainer)</code> on this
 * <code>Transporter</code> to send an exchange. <br>
 * This <code>Transporter</code> sends the <code>MessageExchange</code> to
 * the distant <code>Transporter</code>. <br>
 * When the distant <code>Transporter</code> receives the
 * <code>MessageExchange</code>, it has to call
 * <code>receive(exchange)</code> on its local <code>Router</code>. <br>
 * The localization of the distant container is in charge of the
 * <code>Transporter</code>. <br>
 * <br>
 * <em>Note that the destination container can be the same than this one.</em>
 * 
 * @version $Rev: 477 $ $Date: 2006-05-29 17:18:07 +0200 (lun., 29 mai 2006) $
 * @since Petals 1.0
 * @author alouis
 * @author rnaudin
 */
public interface Transporter extends PetalsLifeCycle {

    /**
     * Constant used to specify the property name to set the transport policy.
     * Used only in plaform distribution. The default policy is the reliability.
     */
    static final String PROPERTY_TRANSPORT_QOS = "org.ow2.petals.transport.qos";

    /**
     * Constant used to specify the reliable policy. It assures the persistence
     * of the exchange. Message are never lost, but there is effort on the
     * duration of the exchange
     */
    static final String RELIABLE_POLICY = "reliable";

    /**
     * Constant used to specify the fast policy. It assures the fastest exchange
     * processing. If there is a problem during communication, the message can
     * be lost.
     */
    static final String FAST_POLICY = "fast";

    /**
     * The default QOS policy.
     */
    static final String DEFAULT_POLICY = RELIABLE_POLICY;

    /**
     * The timeout of the stop of the Transporter
     */
    static final int STOP_TIMEOUT = 10000;

    /**
     * Send the specified exchange to the specified container. According to the
     * transport policy, the message is may be not persisted by the Transport
     * layer, so it can be lost in case of problem.
     * 
     * @param exchange
     *            the exchange to send. <b>MUST</b> be non null
     * @param componentName
     *            the name of the distant container. <b>MUST</b> be non null
     * @param containerName
     *            the name of the distant container. <b>MUST</b> be non null
     *            and non empty
     * @param timeToLive
     *            duration of the message life (in ms). 0 for no timeout
     * @throws TransportException
     *             If the exchange is invalid, or can not be serialized. If the
     *             distant container can not be found.
     */
    void send(MessageExchangeImpl exchange, String componentName, String containerName,
            long timeToLive) throws TransportException;

    /**
     * Send the specified exchange to the specified container and wait until the
     * response is received or the timeout is reached. According to the
     * transport policy, the message is may be not persisted by the Transport
     * layer, so it can be lost in case of problem.
     * 
     * @param exchange
     *            the exchange to send. <b>MUST</b> be non null
     * @param componentName
     *            the name of the distant container. <b>MUST</b> be non null
     * @param containerName
     *            the name of the distant container. <b>MUST</b> be non null
     *            and non empty
     * @param timeToLive
     *            duration of the message life (in ms). 0 for no timeout
     * @throws TransportException
     *             If the exchange is invalid, or can not be serialized. If the
     *             distant container can not be found.
     */
    MessageExchangeImpl sendSync(MessageExchangeImpl exchange, String componentName,
            String containerName, long timeToLive) throws TransportException;

    /**
     * Inform the <code>Transporter</code> that the specified component is
     * installing. The transporter may need to initialize a device to handle new
     * bound destination.
     * 
     * @param componentName
     *            The unique name of the installed component. <b>MUST</b> be
     *            non null
     * @throws TransportException
     */

    void addDestination(String componentName) throws TransportException;

    /**
     * Inform the <code>Transporter</code> that the specified component is
     * uninstalling. The <code>Transporter</code> may need to cleanup the
     * device that handles the bound destination.
     * 
     * @param componentName
     *            The unique name of the uninstalled component. <b>MUST</b> be
     *            non null
     * @throws TransportException
     */

    void removeDestination(String componentName) throws TransportException;

    /**
     * Poll a received message for the given component
     * 
     * @param componentName
     *            The unique name of the component; <b>MUST</b> be non null
     * @param timeoutMS
     *            parameter which represents
     *            <ul>
     *            <li> The time to wait if greater than 0 </li>
     *            <li> A non blocking call if equal to 0</li>
     *            <li> Infinite Blocking call if less than 0</li>
     *            </ul>
     * 
     * @param block
     *            If <code>True</code>, the poll is blocking until a response
     * @return The received message exchange
     */
    MessageExchangeImpl receive(String componentName, long timeoutMS) throws TransportException;

    /**
     * Start the Traffic with the components. this method activates all the
     * communication traffics between components. This method is called once the
     * JBI recovery step is done.
     */
    void startTraffic();

    /**
     * Prepare the stopping of the component. It means that all ongoing message
     * traffic must terminate during the method invocation and all new message
     * traffic is rejected.
     */
    void stopTraffic();

    List<String> getTransporters();

    /**
     * Retrieve the monitor agent for the given protocol.
     * 
     * @param protocolName
     *            a transporter protocol name
     * @return the transporter monitor. Can be {@code null} if no protocol match
     *         the given name
     * @throws TransportException
     *             if an error occurs during retrieve of the protocol monitor
     */
    ProtocolMonitor getProtocolMonitor(String protocolName) throws TransportException;
}
