/**
 * PETALS: PETALS Services Platform
 * Copyright (C) 2005 EBM WebSourcing
 *
 * This library is free software; you can redistribute it and/or
 * modify it under the terms of the GNU Lesser General Public
 * License as published by the Free Software Foundation; either
 * version 2.1 of the License, or any later version.
 *
 * This library is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 * Lesser General Public License for more details.
 *
 * You should have received a copy of the GNU Lesser General Public
 * License along with this library; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307
 * USA.
 *
 * Initial developer(s): pierre.garcia@inrialpes.fr
 *
 */

package org.ow2.petals.transport.platform;

import java.util.Collection;
import java.util.ConcurrentModificationException;
import java.util.HashMap;
import java.util.Hashtable;
import java.util.Iterator;
import java.util.LinkedList;
import java.util.List;
import java.util.Map;
import java.util.Vector;
import java.util.concurrent.BlockingQueue;
import java.util.concurrent.LinkedBlockingQueue;
import java.util.concurrent.TimeUnit;

import javax.jbi.JBIException;
import javax.jbi.messaging.MessageExchange;
import javax.jbi.messaging.MessagingException;

import org.objectweb.fractal.fraclet.annotation.annotations.FractalComponent;
import org.objectweb.fractal.fraclet.annotation.annotations.Interface;
import org.objectweb.fractal.fraclet.annotation.annotations.LifeCycle;
import org.objectweb.fractal.fraclet.annotation.annotations.Monolog;
import org.objectweb.fractal.fraclet.annotation.annotations.Provides;
import org.objectweb.fractal.fraclet.annotation.annotations.Requires;
import org.objectweb.fractal.fraclet.annotation.annotations.type.Cardinality;
import org.objectweb.fractal.fraclet.annotation.annotations.type.LifeCycleType;
import org.objectweb.util.monolog.api.Logger;
import org.ow2.petals.jbi.messaging.endpoint.AbstractEndpoint;
import org.ow2.petals.jbi.messaging.exchange.MessageExchangeImpl;
import org.ow2.petals.transport.TransportException;
import org.ow2.petals.transport.Transporter;
import org.ow2.petals.transport.platform.dream.DreamWrapper;
import org.ow2.petals.transport.platform.joram.JoramTransporter;
import org.ow2.petals.transport.platform.memory.MemoryTransporter;
import org.ow2.petals.transport.util.TransporterUtil;
import org.ow2.petals.kernel.configuration.ConfigurationService;
import org.ow2.petals.kernel.configuration.ContainerConfiguration;
import org.ow2.petals.util.LoggingUtil;

/**
 * This class represents a dispatcher used to transmit messages to the right
 * transporter. When a message arrives, it is sent the message to the
 * appropriate transportProtocol.<br>
 * <br>
 * The transportProtocol to use can be specified by two means. By increasing
 * order, here are the means :
 * <ol>
 * <li> To set the <code>transporter</code> in the <em>server.properties</em>
 * configuration file</li>
 * <li>To set the {@link Transporter.PROPERTY_PROTOCOLS_TO_TRY} in the
 * {@link MessageExchangeImpl message exchange}</li>
 * </ol>
 * In both cases, this property can be a single value or a list of value
 * separated by comas <br>
 * <br>
 *
 * @author pgarcia
 */
@FractalComponent
@Provides(interfaces = {
        @Interface(name = "service", signature = org.ow2.petals.transport.Transporter.class),
        @Interface(name = "dispatcher", signature = org.ow2.petals.transport.platform.TransportProtocolListener.class) })
public class DispatcherImpl implements Transporter, TransportProtocolListener {

    // --------------------------------------------------------------------------
    // Constants
    // --------------------------------------------------------------------------

    /**
     * Interface name (collection) of the protocols available
     */
    public static final String DELEGATED_ITF_NAME = "delegated-service";

    private static HashMap<String, LinkedList<String>> PROTOCOL_POLICIES_MAP = new HashMap<String, LinkedList<String>>();

    public static LinkedList<String> allTransportProtocolsLinkedList = new LinkedList<String>();

    static {
        LinkedList<String> reliabilityLinkedList = new LinkedList<String>();
        reliabilityLinkedList.add(DispatcherImpl.DELEGATED_ITF_NAME + "-"
                + JoramTransporter.TRANSPORTER_NAME);
        PROTOCOL_POLICIES_MAP.put(Transporter.RELIABLE_POLICY, reliabilityLinkedList);

        LinkedList<String> fastLinkedList = new LinkedList<String>();
        fastLinkedList.add(DispatcherImpl.DELEGATED_ITF_NAME + "-"
                + MemoryTransporter.TRANSPORTER_NAME);
        fastLinkedList.add(DispatcherImpl.DELEGATED_ITF_NAME + "-" + DreamWrapper.TRANSPORTER_NAME);
        fastLinkedList.add(DispatcherImpl.DELEGATED_ITF_NAME + "-"
                + JoramTransporter.TRANSPORTER_NAME);
        PROTOCOL_POLICIES_MAP.put(Transporter.FAST_POLICY, fastLinkedList);

        allTransportProtocolsLinkedList.add(DispatcherImpl.DELEGATED_ITF_NAME + "-"
                + MemoryTransporter.TRANSPORTER_NAME);
        allTransportProtocolsLinkedList.add(DispatcherImpl.DELEGATED_ITF_NAME + "-"
                + DreamWrapper.TRANSPORTER_NAME);
        allTransportProtocolsLinkedList.add(DispatcherImpl.DELEGATED_ITF_NAME + "-"
                + JoramTransporter.TRANSPORTER_NAME);
    }

    // --------------------------------------------------------------------------
    // Loggers
    // --------------------------------------------------------------------------

    /**
     * Logger
     */
    @Monolog(name = "logger")
    protected Logger logger;

    /**
     * Logger Wrapper
     */
    protected LoggingUtil log;

    // --------------------------------------------------------------------------
    // Fields
    // --------------------------------------------------------------------------

    /**
     * The local container information.
     */
    protected ContainerConfiguration containerConfiguration;

    /**
     * Map which will be modified by transporters when a normal message is
     * available for a given component
     */
    protected Map<String, BlockingQueue<TransportProtocol>> availableExchangesMap;

    /**
     * Map of MsgExchange that are blocked during a synchronous send
     */
    protected Map<String, MessageExchangeImpl> pendingExchangesMap;

    /**
     * list of default transport protocols to use when a message exchange does
     * not specified the transport protocols to use.
     *
     * values contained in this list MUST match an existing transport protocol.
     */
    protected List<String> defaultProtocolsList;

    /**
     * The list of the threads which are blocking on a poll
     */
    private Vector<Thread> threadsList;

    /**
     * Traffic flag to handle stop of traffic
     */
    private boolean stopTraffic;

    /**
     * Traffic flag to handle start of traffic
     */
    private Boolean[] startTraffic;

    // --------------------------------------------------------------------------
    // Fractal Client Interfaces
    // --------------------------------------------------------------------------

    /**
     * The Configuration service
     */
    @Requires(name = "configuration", signature = org.ow2.petals.kernel.configuration.ConfigurationService.class)
    private ConfigurationService configurationService;

    /**
     * Services to which the transport is delegated
     */
    @Requires(name = DELEGATED_ITF_NAME, signature = TransportProtocol.class, cardinality = Cardinality.COLLECTION)
    protected Map<String, Object> protocols = new HashMap<String, Object>();

    // --------------------------------------------------------------------------
    // Transporter Interface Implementation
    // --------------------------------------------------------------------------

    /*
     * (non-Javadoc)
     *
     * @see org.ow2.petals.transport.Transporter#addDestination(java.lang.String)
     */
    public void addDestination(String componentName) throws TransportException {
        this.log.start();
        // create an entry in message available
        this.availableExchangesMap.put(componentName, new LinkedBlockingQueue<TransportProtocol>());
        // add the destination to each transporter
        Iterator<String> it = this.protocols.keySet().iterator();
        boolean exceptionRaised = false;
        TransportProtocol trProtocol;
        String trProtocolName;
        while (it.hasNext()) {
            trProtocolName = it.next();
            trProtocol = (TransportProtocol) this.protocols.get(trProtocolName);
            try {
                trProtocol.addDestination(componentName);
            } catch (TransportException e) {
                exceptionRaised = true;
                this.log.error("Cannot add the destination '" + componentName
                        + "' to the transport protocol " + trProtocolName);
            }
        }
        if (exceptionRaised)
            throw new TransportException("An error occured while adding the destination "
                    + componentName);
        this.log.end();
    }

    /*
     * (non-Javadoc)
     *
     * @see org.ow2.petals.transport.Transporter#removeDestination(java.lang.String)
     */
    public void removeDestination(final String componentName) throws TransportException {
        this.log.start();
        // remove the destination from the messageAvailable
        this.availableExchangesMap.remove(componentName);

        // TODO manage remaining messages in the waitingExchange queue
        // and transporter name in the messageAvailable map.

        // remove the destination from the transport protocols
        Iterator<String> it = this.protocols.keySet().iterator();
        boolean exceptionRaised = false;
        TransportProtocol trProtocol;
        String trProtocolName;
        while (it.hasNext()) {
            trProtocolName = it.next();
            trProtocol = (TransportProtocol) this.protocols.get(trProtocolName);
            try {
                trProtocol.removeDestination(componentName);
            } catch (TransportException e) {
                exceptionRaised = true;
                this.log.error("Cannot remove the destination '" + componentName
                        + "' from the transport protocol " + trProtocolName);
            }
        }
        if (exceptionRaised)
            throw new TransportException("An error occured while removing the destination '"
                    + componentName + "'");
        this.log.end();
    }

    /*
     * (non-Javadoc)
     *
     * @see org.ow2.petals.transport.Transporter#sendSync(org.ow2.petals.jbi.messaging.exchange.MessageExchangeImpl,
     *      java.lang.String, java.lang.String, long)
     */
    public MessageExchangeImpl sendSync(MessageExchangeImpl exchange, final String componentName,
            final String containerName, final long timeToLive) throws TransportException {
        this.log.start();
        MessageExchangeImpl responseExchange = null;
        long startTime = 0;

        // clean unused messages
        exchange.cleanMessages();

        // if the dispatcher is stopped, interrupt the current thread
        checkTransporterState();

        this.log.debug("Dispatching a transfer of message to component '" + componentName
                + "' on container '" + containerName + "'");

        // set the relevant properties to mark the message as a synchronous send
        try {
            TransporterUtil.setSendSyncProperties(exchange);
        } catch (MessagingException e) {
            throw new TransportException(e);
        }

        // mark the start time to handle the timeout
        if (timeToLive > 0) {
            startTime = System.currentTimeMillis();
        }

        LinkedList<String> listProtocols = this.getTransportProtocolNames(exchange, containerName);

        MultiTransportException multiTransportException = null;
        for (String protocolName : listProtocols) {

            this.log.debug("Try to send via the protocol " + protocolName + " with a timeout of "
                    + timeToLive + " milliseconds");
            TransportProtocol protocol = null;

            try {
                protocol = this.getProtocol(protocolName);
            } catch (TransportException e) {
                this.log.error(e.getMessage());
                continue;
            }

            if (timeToLive > 0) {
                // If the timeout is reached, we return null
                if (System.currentTimeMillis() > (startTime + timeToLive)) {
                    this.log.warning("Timeout reached for synchronous send of the exchange: "
                            + exchange.getExchangeId());
                    this.log.end();
                    return null;
                }
            }

            try {
                responseExchange = this.sendSyncViaProtocol(exchange, componentName, containerName,
                        timeToLive, protocol);
            } catch (TransportException e) {
                this.log.warning("The synchronous send with the protocol " + protocolName
                        + " threw a " + TransportException.class.getName(), e);
                if (multiTransportException == null) {
                    multiTransportException = new MultiTransportException(
                            "The synchronous send failed for multiple reasons. "
                                    + "See PEtALS logs or StackTrace for more information", e);
                } else {
                    multiTransportException.addCount();
                }
                continue;
            }

            // a sendSync may have return null because the dispatcher is
            // stopping
            if (responseExchange == null) {
                checkTransporterState();
            }

            break;
        }

        if (responseExchange == null) {
            if (multiTransportException != null
                    && listProtocols.size() == multiTransportException.getExceptionCount()) {
                throw multiTransportException;
            } else {
                this.log.error("Failed to send synchronously the exchange: "
                        + exchange.getExchangeId() + ". Timeout occured");
            }
        }

        this.log.end();
        return responseExchange;
    }

    /**
     * Send the message via the specified TransportProtocol(s). the property
     * {@link Transporter#PROPERTY_TRANSPORT_QOS} can be set in the
     * <code>messageExchange</code> to specify the transporter to use,
     * otherwise the {@link Transporter#DEFAULT_POLICY} is used Currently
     * available transporters are :
     * <ul>
     * <li>joram</li>
     * <li>dream</li>
     * <li>lcoal</li>
     * </ul>
     *
     * @see org.ow2.petals.transport.Transporter#send(org.ow2.petals.jbi.messaging.exchange.MessageExchangeImpl,
     *      java.lang.String, java.lang.String, long)
     */
    public void send(MessageExchangeImpl exchange, String componentName, String containerName,
            long timeToLive) throws TransportException {
        this.log.start();
        boolean isSent = false;

        // clean unused messages
        exchange.cleanMessages();

        // if the dispatcher is stopped, interrupt the current thread
        checkTransporterState();

        this.log.debug("Dispatching a transfer of message to component '" + componentName
                + "' on container '" + containerName + "'");

        LinkedList<String> listProtocols = this.getTransportProtocolNames(exchange, containerName);

        TransportProtocol protocol = null;
        for (String protocolName : listProtocols) {
            // skip the Memory Protocol if the destination is not local
            if (protocolName.equals(DispatcherImpl.DELEGATED_ITF_NAME + "-"
                    + MemoryTransporter.TRANSPORTER_NAME)
                    && !containerName.equals(this.containerConfiguration.getName())) {
                continue;
            }

            protocol = this.getProtocol(protocolName);

            this.log.debug("Try to send with the protocol " + protocolName + " and a timeout of "
                    + timeToLive + " milliseconds");

            try {
                protocol.send(exchange, componentName, containerName, timeToLive);
            } catch (TransportException e) {
                this.log.warning("The asynchronious send with the protocol " + protocolName
                        + " threw a " + TransportException.class.getName(), e);
                continue;
            }
            isSent = true;
            break;
        }

        if (isSent == false) {
            throw new TransportException("Failed to send asynchronously the exchange: "
                    + exchange.getExchangeId());
        }

        this.log.end();
    }

    /*
     * (non-Javadoc)
     *
     * @see org.ow2.petals.transport.Transporter#receive(java.lang.String,
     *      long)
     */
    public MessageExchangeImpl receive(String componentName, long timeoutMS)
            throws TransportException {
        this.log.start();

        synchronized (this.startTraffic) {
            if (this.startTraffic[0].equals(false)) {
                // wait until the start of the traffics
                try {
                    this.startTraffic.wait();
                } catch (InterruptedException e) {
                    throw new TransportException("The pending receipt for the component '"
                            + componentName + "' is interrupted", e);
                }
            }
        }

        // if the dispatcher is stopped, interrupt the current thread
        checkTransporterState();

        this.log.debug("Poll a message for the component '" + componentName
                + "' with a timeout of " + timeoutMS);

        BlockingQueue<TransportProtocol> matchingQueue = this.availableExchangesMap
                .get(componentName);
        if (matchingQueue == null)
            throw new TransportException("The component '" + componentName
                    + "' destination doesn't exists");

        TransportProtocol transportProtocol;

        if (timeoutMS == 0) {
            // non-blocking call
            transportProtocol = matchingQueue.poll();
        } else {
            this.threadsList.add(Thread.currentThread());
            try {
                if (timeoutMS > 0) {
                    // blocking call with timeout
                    transportProtocol = matchingQueue.poll(timeoutMS, TimeUnit.MILLISECONDS);
                } else {
                    // blocking call without timeout
                    transportProtocol = matchingQueue.take();
                }
            } catch (InterruptedException e) {
                Thread.currentThread().interrupt();
                throw new TransportException("The pending receipt for the component '"
                        + componentName + "' is interrupted", e);
            } finally {
                this.threadsList.remove(Thread.currentThread());
            }
        }

        if (transportProtocol == null) {
            return null;
        }
        MessageExchangeImpl msgExchange = transportProtocol.receive(componentName);

        this.log.end();
        return msgExchange;
    }

    /*
     * (non-Javadoc)
     *
     * @see org.ow2.petals.transport.Transporter#startTraffic()
     */
    public void startTraffic() {
        // wake up all the pending receive waiting the start of traffics
        synchronized (this.startTraffic) {
            this.startTraffic[0] = Boolean.valueOf(true);
            this.startTraffic.notifyAll();

        }
    }

    /*
     * (non-Javadoc)
     *
     * @see org.ow2.petals.transport.Transporter#prepareStop()
     */
    public void stopTraffic() {
        this.log.start();
        boolean redo = true;
        this.stopTraffic = true;

        // Wait that all the synchronous invocation finished or the timeout is
        // reached
        int lapsTime = 0;
        try {
            while (lapsTime < Transporter.STOP_TIMEOUT) {
                lapsTime += 1000;

                if (this.pendingExchangesMap.size() > 0) {
                    // sleep for 1 second
                    Thread.sleep(1000);
                } else {
                    break;
                }
            }
        } catch (InterruptedException e) {
            // do nothing
        }

        if (lapsTime >= STOP_TIMEOUT) {
            this.log.warning("The timeout for stopping the Transporter has been reached."
                    + " Some synchronous invocation are lost.");
        }

        /*
         * Awake all the pending 'SYNC_SEND' issuer threads to abort the
         * invocation
         */
        while (redo) {
            try {
                Collection<MessageExchangeImpl> values = this.pendingExchangesMap.values();
                for (MessageExchangeImpl messageExchangeImpl : values) {
                    synchronized (messageExchangeImpl) {
                        messageExchangeImpl.notify();
                    }
                }
                redo = false;
            } catch (ConcurrentModificationException e) {
                // Try again
            }
        }

        redo = true;
        /*
         * Interrupt all the pending receive threads
         */
        while (redo) {
            try {
                for (Thread thread : this.threadsList) {
                    thread.interrupt();
                }
                redo = false;
            } catch (ConcurrentModificationException e) {
                // Try again
            }
        }

        this.log.end();
    }

    /*
     * (non-Javadoc)
     *
     * @see org.ow2.petals.transport.Transporter#getTransporters()
     */
    public List<String> getTransporters() {
        return DispatcherImpl.allTransportProtocolsLinkedList;
    }

    /*
     * (non-Javadoc)
     *
     * @see org.ow2.petals.transport.Transporter#ProtocolMonitor(java.lang.String)
     */
    public org.ow2.petals.transport.ProtocolMonitor getProtocolMonitor(
            String protocolName) throws TransportException {

        TransportProtocol transportProtocol = getProtocol(protocolName);

        return transportProtocol.getProtocolMonitor();
    }

    // --------------------------------------------------------------------------
    // PetalsLifeCycle Interface Implementation
    // --------------------------------------------------------------------------

    /*
     * (non-Javadoc)
     *
     * @see org.objectweb.petals.PetalsLifeCycle#shutdown()
     */
    public void shutdown() throws Exception {
        this.log.start();
        Iterator<String> it = this.protocols.keySet().iterator();
        boolean exceptionRaised = false;
        TransportProtocol trProtocol;
        String trProtocolName;
        while (it.hasNext()) {
            trProtocolName = it.next();
            trProtocol = (TransportProtocol) this.protocols.get(trProtocolName);
            try {
                trProtocol.shutdown();
            } catch (Exception e) {
                exceptionRaised = true;
                this.log.error("Cannot shut down the transport protocol " + trProtocolName);
            }
        }
        if (exceptionRaised)
            throw new Exception("An error occured while the dispatcher was shutdown ");
        this.log.end();
    }

    // --------------------------------------------------------------------------
    // TransportProtocolListener interface implementation
    // --------------------------------------------------------------------------

    /*
     * (non-Javadoc)
     *
     * @see org.ow2.petals.transport.platform.TransportProtocolListener#messageAvailable(org.ow2.petals.transport.platform.TransportProtocol,
     *      java.lang.String)
     */
    public void messageAvailable(TransportProtocol transportProtocol, String componentName) {
        this.log.start();

        BlockingQueue<TransportProtocol> matchingQueue = this.availableExchangesMap
                .get(componentName);
        if (matchingQueue == null) {
            this.log.error("The blocking queue for the component " + componentName
                    + " must not be null");
        } else {
            try {
                matchingQueue.put(transportProtocol);
            } catch (InterruptedException e) {
                e.printStackTrace();
            }
        }

        this.log.end();
    }

    /*
     * (non-Javadoc)
     *
     * @see org.ow2.petals.transport.platform.TransportProtocolListener#syncMessageReceived(org.ow2.petals.jbi.messaging.exchange.MessageExchangeImpl)
     */
    public void syncMessageReceived(MessageExchangeImpl exchange) {
        this.log.start();

        if (!this.onMessageExchange(exchange)) {
            this.sendBackErrorExchange(exchange);
        }

        this.log.end();
    }

    // --------------------------------------------------------------------------
    // Fractal Lifecycle methods
    // --------------------------------------------------------------------------
    /**
     * Initialize variables.
     *
     * @throws Exception
     */
    @LifeCycle(on = LifeCycleType.START)
    protected void start() throws Exception {
        this.log = new LoggingUtil(this.logger);
        this.log.start();
        this.startTraffic = new Boolean[1];
        this.startTraffic[0] = Boolean.valueOf(false);
        this.availableExchangesMap = new Hashtable<String, BlockingQueue<TransportProtocol>>();
        this.pendingExchangesMap = new Hashtable<String, MessageExchangeImpl>();
        this.threadsList = new Vector<Thread>();
        this.containerConfiguration = this.configurationService.getContainerConfiguration();
        this.log.end();
    }

    @LifeCycle(on = LifeCycleType.STOP)
    protected void stop() throws Exception {
        this.log.call();
    }

    /**
     * Utility to get the matching Transport Protocol component
     *
     * @param protocolName
     *            The name of the Transport Protocol
     */
    private TransportProtocol getProtocol(String protocolName) throws TransportException {
        this.log.start("Transport protocol name : " + protocolName);
        TransportProtocol protocol = (TransportProtocol) this.protocols.get(protocolName);
        if (protocol == null)
            throw new TransportException("The transport protocol linked to the interface '"
                    + protocolName + "' does not exist");
        this.log.end();
        return protocol;
    }

    /**
     * Receive a response message from a synchronous send.
     *
     * @param exchange
     *            The received message exchange
     * @return <code>true</code> if the response has been correlated with the
     *         issuer, <code>false</code> otherwise
     */
    private boolean onMessageExchange(MessageExchangeImpl exchange) {
        this.log.start();

        boolean isMatched = false;

        // update the message properties
        TransporterUtil.updateSyncProperties(exchange);

        // push the response message exchange
        if (this.pushSyncResponse(exchange)) {
            isMatched = true;
        } else {
            // set an error ack to the exchange
            exchange.setError(new MessagingException(
                    "Reject a message Exchange response from a previous synchronious send"));
            this.log
                    .warning("Reject a message Exchange response from a previous synchronious send. ExchangeId: "
                            + exchange.getExchangeId());
        }
        this.log.end();
        return isMatched;
    }

    /**
     * Send synchronously the message exchange via the protocol given as
     * parameter.
     *
     * @param exchange
     * @param componentName
     * @param containerName
     * @param timeToLive
     * @param protocol
     * @return
     * @throws TransportException
     */
    private MessageExchangeImpl sendSyncViaProtocol(MessageExchangeImpl exchange,
            String componentName, String containerName, long timeToLive, TransportProtocol protocol)
            throws TransportException {
        this.log.start();

        // Reference this exchange in the Map of waitingExchanges.
        // It will be used to wake up the exchange when the response
        // is arrived
        this.pendingExchangesMap.put(TransporterUtil.waitingExchangeKey(exchange), exchange);

        this.log.debug("The message will be sent asynchronously via the given protocol");

        try {
            synchronized (exchange) {
                // Send the message
                protocol.send(exchange, componentName, containerName, timeToLive);
                // wait for response or a timeout reached
                exchange.wait(timeToLive);
            }
        } catch (InterruptedException e) {
            this.pendingExchangesMap.remove(TransporterUtil.waitingExchangeKey(exchange));
            throw new TransportException(
                    "Failed to receive the response message of the synchronious send", e);
        } catch (TransportException e) {
            this.pendingExchangesMap.remove(TransporterUtil.waitingExchangeKey(exchange));
            throw e;
        }

        // Now, the response has been sent (wait unblocked),
        // or timeout occured.

        // get the exchange from the map
        MessageExchangeImpl responseExchange = this.pendingExchangesMap.remove(TransporterUtil
                .waitingExchangeKey(exchange));

        // If the exchange is not the same than the sent exchange,
        // it means that a response message has been received.

        // If it is the same but synchronous property is not set,
        // it means that it is a response from the Memory transporter.

        // Otherwise, it means that the timeout has been reached;
        // the response exchange is set to null
        if (responseExchange == exchange && TransporterUtil.getSyncMode(responseExchange, false)) {
            responseExchange = null;
        }
        this.log.end();
        return responseExchange;
    }

    /**
     * Send back to the issuer of the response an error ack message
     *
     * @param exchange
     *            the message to send back
     */
    private void sendBackErrorExchange(MessageExchangeImpl exchange) {
        this.log.start();
        // set the timeout arbitrary to 10 seconds
        try {
            AbstractEndpoint endpoint = null;
            if (MessageExchange.Role.PROVIDER.equals(exchange.getRole())) {
                exchange.setRole(MessageExchange.Role.CONSUMER);
                endpoint = exchange.getConsumerEndpoint();
            } else {
                exchange.setRole(MessageExchange.Role.PROVIDER);
                endpoint = (AbstractEndpoint) exchange.getEndpoint();
            }
            this.send(exchange, endpoint.getComponentName(), endpoint.getContainerName(), 10000L);
        } catch (TransportException e) {
            // do nothing, the message is lost.
        }
        this.log.end();
    }

    /**
     * The message is a response of a sendSync() method, the matching sendSync()
     * must be unblocked with the attached response.
     *
     * @param msg
     *            the message received by the <code>Transporter</code>
     * @throws JBIException
     *             The channel has been closed by the component. The message can
     *             not be accepted
     */
    private boolean pushSyncResponse(MessageExchangeImpl responseExchange) {
        this.log.start();
        boolean isMessageMatched = false;

        // Check if this ME is synchronized to a blocked synchronous send.
        // If so, we unblock the sendSync() method, otherwise,
        // we return the message to the local topic, to be consumed by the next
        // accept().

        MessageExchangeImpl sentExchange = this.pendingExchangesMap.get(TransporterUtil
                .waitingExchangeKey(responseExchange));

        if (sentExchange != null) {
            this.pendingExchangesMap.put(TransporterUtil.waitingExchangeKey(responseExchange),
                    responseExchange);
            isMessageMatched = true;
            // we notify the exchange issuer that its response is received.
            synchronized (sentExchange) {
                sentExchange.notify();
            }
        }
        this.log.end();

        return isMessageMatched;
    }

    // --------------------------------------------------------------------------
    // Utility methods
    // --------------------------------------------------------------------------

    /**
     * Get the list of Transport Protocol names to use according to the given
     * exchange.
     *
     * @param exchange
     *            The exchange message to transfer
     *
     * @return the list of Transport Protocols
     * @throws TransportException
     */
    private LinkedList<String> getTransportProtocolNames(MessageExchangeImpl exchange,
            String containerName) throws TransportException {
        this.log.start();
        LinkedList<String> listProtocols = null;

        String transportPolicy = (String) exchange.getProperty(Transporter.PROPERTY_TRANSPORT_QOS);

        // Get the default policy if null and set the property in the message
        if (transportPolicy == null) {
            transportPolicy = this.containerConfiguration.getTransportQOS();
            exchange.setProperty(Transporter.PROPERTY_TRANSPORT_QOS, transportPolicy);
        } else {
            if (!PROTOCOL_POLICIES_MAP.containsKey(transportPolicy)) {
                throw new TransportException(
                        "The transport QOS specified in the message exchange is not valid: "
                                + transportPolicy);
            }
        }

        listProtocols = new LinkedList<String>(PROTOCOL_POLICIES_MAP.get(transportPolicy));

        // remove the unused Transporter Protocol
        if (Transporter.FAST_POLICY.equals(transportPolicy)) {
            if (containerName.equals(this.containerConfiguration.getName())) {
                listProtocols.remove(DispatcherImpl.DELEGATED_ITF_NAME + "-"
                        + DreamWrapper.TRANSPORTER_NAME);
            } else {
                listProtocols.remove(DispatcherImpl.DELEGATED_ITF_NAME + "-"
                        + MemoryTransporter.TRANSPORTER_NAME);
            }
        }

        this.log.end("Process the '" + transportPolicy + "' QOS");
        return listProtocols;
    }

    /**
     * Check if the transporter is running, if not, raise an exception.
     *
     * @throws TransportException
     */
    private void checkTransporterState() throws TransportException {
        this.log.call();

        if (this.stopTraffic) {
            Thread.currentThread().interrupt();
            throw new TransportException("The Transporter is stopped", new InterruptedException());
        }
    }

}
