/**
 * PETALS - PETALS Services Platform.
 * Copyright (c) 2005 EBM Websourcing, http://www.ebmwebsourcing.com/
 *
 * This library is free software; you can redistribute it and/or
 * modify it under the terms of the GNU Lesser General Public
 * License as published by the Free Software Foundation; either
 * version 2.1 of the License, or (at your option) any later version.
 * This library is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 * Lesser General Public License for more details.
 *
 * You should have received a copy of the GNU Lesser General Public
 * License along with this library; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
 *
 * -------------------------------------------------------------------------
 * $Id: JoramConfigurationBuilder.java 2:04:08 PM ddesjardins $
 * -------------------------------------------------------------------------
 */
package org.ow2.petals.transport.util;

import java.io.File;
import java.io.IOException;
import java.util.Set;

import javax.xml.bind.JAXBException;

import org.objectweb.joram.client.jms.admin.AdminModule;
import org.ow2.petals.topology.TopologyBuilder;
import org.ow2.petals.topology.TopologyException;
import org.ow2.petals.topology.generated.Topology;
import org.ow2.petals.PetalsException;
import org.ow2.petals.transport.platform.joram.JoramTransporter;
import org.ow2.petals.kernel.configuration.ContainerConfiguration;
import org.ow2.petals.util.SystemUtil;

/**
 * This class is used to handle the JOrAM topology.
 *
 * @author Roland NAUDIN - eBMWebsourcing
 * @since Petals 2.0
 */
public class JoramTopology {

	/**
	 * The a3servers resource file path
	 */
	private static final String A3SERVER_FILE_NAME = "a3servers.xml";

	/**
	 * Singleton
	 */
	private static JoramTopology instance;

	/**
	 * The topology builder
	 */
	private static TopologyBuilder topologyBuilder;

	/**
	 * The JOrAM control singleton
	 */
	private JoramControl joramControl = JoramControl.getInstance();

	/**
	 * Creates a new instance of {@link JoramTopology}
	 *
	 * @throws TopologyException
	 * @throws JAXBException
	 *
	 */
	private JoramTopology() throws JAXBException, TopologyException {
		topologyBuilder = TopologyBuilder.getInstance();
	}

	/**
	 * Get the instance of the configuration builder
	 *
	 * @return
	 * @throws TopologyException
	 * @throws JAXBException
	 */
	public static synchronized JoramTopology getInstance()
			throws JAXBException, TopologyException {
		if (instance == null) {
			instance = new JoramTopology();
		}
		return instance;
	}

	/**
	 * Inform the running JOrAM server nodes about the presence of the given new
	 * server node.<br>
	 * Try to connect to each JOrAM node to get configuration, then notify them
	 * if necessary about the new JOrAM node.
	 *
	 * @param domainname
	 * @param localContainerConfiguration
	 * @param containersConfiguration
	 */
	public void addServerToTopology(String domainName,
			ContainerConfiguration localContainerConfiguration,
			Set<ContainerConfiguration> containersConfiguration) {

		for (ContainerConfiguration remoteContainerConfiguration : containersConfiguration) {
			if (!remoteContainerConfiguration.getName().equals(
					localContainerConfiguration.getName())) {
				try {
					AdminModule.connect(remoteContainerConfiguration.getHost(),
							remoteContainerConfiguration.getJoramTCPPort(),
							remoteContainerConfiguration.getUser(),
							remoteContainerConfiguration.getPassword(), 3);

					// If the new server has not been already recorded in
					// the remote server topology, add it
					// if (!isServerRegistered(localContainerConfiguration
					// .getJoramId())) {
					System.out.println("add the Joram server "
							+ localContainerConfiguration.getName());
					joramControl.addJoramServer(Integer
							.parseInt(localContainerConfiguration.getName()),
							localContainerConfiguration.getHost(), domainName,
							localContainerConfiguration.getJoramDomainPort(),
							localContainerConfiguration.getName(),
							localContainerConfiguration.getJoramTCPPort());
					// }
					AdminModule.disconnect();
					break;
				} catch (Exception e) {
					// continue the loop
				}
			}
		}
	}

	/**
	 * Remove the local JOrAM server from the JOrAM topology
	 *
	 * @throws PetalsException
	 */
	public void removeServerfromTopology(
			ContainerConfiguration localContainerConfiguration,
			Set<ContainerConfiguration> containersConfiguration)
			throws PetalsException {
		if (containersConfiguration.size() > 1) {
			try {
				AdminModule.collocatedConnect(localContainerConfiguration
						.getUser(), localContainerConfiguration
						.getPassword());
				for (ContainerConfiguration remoteContainerConfiguration : containersConfiguration) {
					if (!remoteContainerConfiguration
							.equals(localContainerConfiguration)) {
						joramControl
								.removeContainer(remoteContainerConfiguration);
					}
				}
				AdminModule.disconnect();

				for (ContainerConfiguration remoteContainerConfiguration : containersConfiguration) {
					if (!remoteContainerConfiguration
							.equals(localContainerConfiguration)) {
						AdminModule
								.connect(
										remoteContainerConfiguration.getHost(),
										remoteContainerConfiguration
												.getJoramTCPPort(),
										remoteContainerConfiguration
												.getUser(),
										remoteContainerConfiguration
												.getPassword(), 3);
						joramControl
								.removeContainer(localContainerConfiguration);
						AdminModule.disconnect();
					}
				}
			} catch (Exception e) {
				// continue the loop
			}

		}
	}

	/**
	 * Build the local JOrAM a3xserver XML file.
	 *
	 * @param topology
	 * @throws IOException
	 * @throws TopologyException
	 */
	public void buildJoramA3ServerFile(Topology topology) throws IOException,
			TopologyException {

		topologyBuilder.writeA3ServerForJoram(topology, SystemUtil
				.getPetalsInstallDirectory()
				+ File.separator
				+ JoramTransporter.TRANSPORTER_NAME
				+ File.separator + A3SERVER_FILE_NAME);
	}
}
